# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains formatter methods for operations related to data
interpolation.
"""

import numpy
import numpy.typing
import pyproj
import typing
import logging
import scipy.interpolate
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


class MissingMandatoryVariable(ValueError):
    """Error raised when at least one variable required by the interpolation
    method is missing from the variables provided as input."""
    pass


class ReferenceVariableMissing(ValueError):
    """Error raised when the variables dictionary does not contain the
    reference variable."""
    def __init__(self, var_name: str) -> None:
        """"""
        msg = (f'Reference variable "{var_name}" not found in the variables '
               'list')
        super(ReferenceVariableMissing, self).__init__(msg)


class CannotInterpolateRefVariable(ValueError):
    """Error raised when a variable is both a target and a reference for the
    interpolation algorithm."""
    def __init__(self, var_id: str) -> None:
        """"""
        self.var_id = var_id


def fill_1d_gaps(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 granule: Granule,
                 *args: typing.Iterable[typing.Any],
                 targets: typing.Iterable[str],
                 reference: str,
                 gaps_indices: numpy.typing.NDArray,
                 method: str,
                 delta: float) -> FormatterResult:
    """
    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables that must be filled with NaN values where gaps
        have been detected
    reference: str
        Identifier of the variable which has been used to detected gaps
    gaps_indices: numpy.ndarray
        Array indices where gaps have been detected
    method: str
        Interpolation method to use for geographical coordinates
    delta: float
        Sampling expected for the reference variable after interpolation

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been modified to fill the gaps
    """
    var_names = list(granule.vars.keys())
    has_ref = (reference in var_names)
    has_lon = ('lon' in var_names)
    has_lat = ('lat' in var_names)
    if False in (has_ref, has_lon, has_lat,):
        raise MissingMandatoryVariable(has_ref, has_lon, has_lat)

    if reference in targets:
        raise CannotInterpolateRefVariable(reference)

    interpolated: typing.Dict[str, typing.List[numpy.typing.NDArray]] = {}
    interpolated[reference] = []
    interpolated['lon'] = []
    interpolated['lat'] = []

    for var_name in targets:
        interpolated[var_name] = []

    chunk_slice = slice(0, gaps_indices[0] + 1)
    for var_name in var_names:
        var_chunk = granule.vars[var_name]['array'][chunk_slice]
        interpolated[var_name].append(var_chunk)

    ref = granule.vars[reference]['array']
    lon = granule.vars['lon']['array']
    lat = granule.vars['lat']['array']
    func_lon = scipy.interpolate.interp1d(ref, lon, kind=method)
    func_lat = scipy.interpolate.interp1d(ref, lat, kind=method)

    for i in range(len(gaps_indices)):
        ref_start = ref[gaps_indices[i]]
        ref_stop = ref[gaps_indices[i] + 1]
        _intermediary_steps = (ref_stop - ref_start) / delta
        intermediary_steps = numpy.ceil(_intermediary_steps).astype('int')

        # Interpolate lon, lat, time
        _ref_chunk = numpy.linspace(ref_start, ref_stop,
                                    num=2 + intermediary_steps)
        ref_chunk = _ref_chunk[1:-1]  # only keep intermediary steps
        lon_chunk = func_lon(ref_chunk)
        lat_chunk = func_lat(ref_chunk)
        interpolated[reference].append(ref_chunk)
        interpolated['lon'].append(lon_chunk)
        interpolated['lat'].append(lat_chunk)
        chunk_shape = numpy.shape(ref_chunk)

        # add dummy chunks for each variable
        for var_name in targets:
            var_chunk = numpy.full(chunk_shape, numpy.nan)
            interpolated[var_name].append(var_chunk)

        slice_first = gaps_indices[i] + 1
        if i < (len(gaps_indices) - 1):
            slice_last = gaps_indices[i + 1]
            chunk_slice = slice(slice_first, slice_last + 1)
        else:
            chunk_slice = slice(slice_first, None)

        interpolated[reference].append(ref[chunk_slice])
        interpolated['lon'].append(lon[chunk_slice])
        interpolated['lat'].append(lat[chunk_slice])
        for var_name in targets:
            var_chunk = granule.vars[var_name]['array'][chunk_slice]
            interpolated[var_name].append(var_chunk)

    # Concatenate original and interpolated chunks to replace the input data
    concat = numpy.concatenate
    granule.vars[reference]['array'] = concat(interpolated[reference], axis=0)
    granule.vars['lon']['array'] = concat(interpolated['lon'], axis=0)
    granule.vars['lat']['array'] = concat(interpolated['lat'], axis=0)
    for var_name in targets:
        granule.vars[var_name]['array'] = concat(interpolated[var_name],
                                                 axis=0)
    granule.dims[reference] = len(granule.vars[reference]['array'])

    return input_opts, output_opts, granule


def geoloc_from_gcps(gcplon: numpy.typing.NDArray,
                     gcplat: numpy.typing.NDArray,
                     gcplin: numpy.typing.NDArray,
                     gcppix: numpy.typing.NDArray,
                     lin: numpy.typing.NDArray,
                     pix: numpy.typing.NDArray
                     ) -> typing.Tuple[numpy.typing.NDArray,
                                       numpy.typing.NDArray]:
    """"""
    geod = pyproj.Geod(ellps='WGS84')
    fwd, bwd, dis = geod.inv(gcplon[:, :-1], gcplat[:, :-1],
                             gcplon[:, 1:], gcplat[:, 1:])

    # Find line and column for the top-left corner of the 4x4 GCPs cell which
    # contains the requested locations
    nlin, npix = gcplat.shape
    _gcplin = gcplin[:, 0]
    _gcppix = gcppix[0, :]
    top_line: numpy.typing.NDArray
    top_line = numpy.searchsorted(_gcplin, lin, side='right') - 1
    left_column: numpy.typing.NDArray
    left_column = numpy.searchsorted(_gcppix, pix, side='right') - 1

    # Make sure this line and column remain within the matrix and that there
    # are adjacent line and column to define the bottom-right corner of the 4x4
    # GCPs cell
    top_line = numpy.clip(top_line, 0, nlin - 2)
    bottom_line = top_line + 1
    left_column = numpy.clip(left_column, 0, npix - 2)
    right_column = left_column + 1

    # Compute coordinates of the requested locations in the 4x4 GCPs cell
    line_extent = _gcplin[bottom_line] - _gcplin[top_line]
    column_extent = _gcppix[right_column] - _gcppix[left_column]
    line_rel_pos = (lin - _gcplin[top_line]) / line_extent
    column_rel_pos = (pix - _gcppix[left_column]) / column_extent

    # Compute geographical coordinates of the requested locations projected on
    # the top and bottom lines
    lon1, lat1, _ = geod.fwd(gcplon[top_line, left_column],
                             gcplat[top_line, left_column],
                             fwd[top_line, left_column],
                             dis[top_line, left_column] * column_rel_pos)
    lon2, lat2, _ = geod.fwd(gcplon[bottom_line, left_column],
                             gcplat[bottom_line, left_column],
                             fwd[bottom_line, left_column],
                             dis[bottom_line, left_column] * column_rel_pos)

    # Compute the geographical coordinates of the requested locations projected
    # on a virtual column joining the projected points on the top and bottom
    # lines
    fwd12, bwd12, dis12 = geod.inv(lon1, lat1, lon2, lat2)
    lon, lat, _ = geod.fwd(lon1, lat1, fwd12, dis12 * line_rel_pos)

    return lon, lat



# vim: ts=4:sts=4:sw=4
#
# @date 2019-10-10
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.lib.time
import idf_converter.lib.generic_grid
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'GRID_YX'


class MissingVariablesList(Exception):
    """Error raised when the "variables" input option has not been specified by
    the user."""
    pass


class MissingProjection(Exception):
    """Error raised when the "projection" input option has not been specified
    by the user."""
    pass


class MissingSpatialResolution(Exception):
    """Error raised when the "spatial_resolution" input option has not been
    specified by the user."""
    pass


class MissingCoordinatesSystem(Exception):
    """Error raised when the data coordinates system has not been fully defined
    by the user. At least one of the (y_variable, x_variable) and
    (lat_variable, lon_variable) couples must be specified."""
    pass


class MissingValidTime(Exception):
    """Error raised when time values are read from a variable in the input
    file but none of them is valid (masked values)."""
    def __init__(self, time_var_id: str) -> None:
        """"""
        self.time_var_id = time_var_id


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    variables\tComma-separated list of variable identifiers to'
           ' extract from the input file',
           '    lon_variable\tIdentifier of the longitude variable',
           '    lat_variable\tIdentifier of the latitude variable',
           '    lon_dimension\tIdentifier of the longitude dimension'
           ' (defaults to the value of lon_variable)',
           '    lat_dimension\tIdentifier of the latitude dimension'
           ' (defaults to the value of lat_variable)',
           '    lon_reversed\tSet to yes|true|1 if longitude values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    lat_reversed\tSet to yes|true|1 if latitude values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    x_variable\tIdentifier of the x variable',
           '    y_variable\tIdentifier of the y variable',
           '    x_dimension\tIdentifier of the x dimension'
           ' (defaults to the value of x_variable)',
           '    y_dimension\tIdentifier of the y dimension'
           ' (defaults to the value of y_variable)',
           '    x_reversed\tSet to yes|true|1 if x values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    y_reversed\tSet to yes|true|1 if y values should'
           ' be traversed in reverse order (defaults to no|false|0)',
           '    time_variable\tIdentifier of the time variable (optional).'
           ' If time_variable is not defined, then time_coverage_start'
           ' and time_coverage_end attributes must be defined either in the'
           ' input file or via global_overrides',
           '    depth_variable\tIdentifier of the depth variable (optional).'
           ' If depth_variable is defined, it will be assumed that it'
           ' corresponds to the slowest changed dimension after the time '
           ' dimension',
           '    global_overrides\tDefinition of global attributes that are'
           ' either missing or erroneous in the input file. The value of this'
           ' option is a comma-separated list of key:value',
           '    variable_overrides_VAR\tDefinition of variable attributes'
           ' that are either missing or erroneous in the input file.'
           ' VAR is the identifier of the variable and the value of this'
           ' option is a comma-separated list of key:value',
           '    projection\tDefinition of the projection for the data grid'
           ' (Proj4 format)',
           '    spatial_resolution\tResolution of the grid in meters'
           '    time_coverage_relative_start\tOffset (in seconds) that must'
           ' be applied to the values of the time variable to obtain the lower'
           ' bound of the time coverage. This option is required when there is'
           ' more than one time value in the input file or when it is not'
           ' possible to provide a static value for time_coverage_start and'
           ' time_coverage_end (via input file or global overrides)',
           '    time_coverage_relative_end\tOffset (in seconds) that must'
           ' be applied to the values of the time variable to obtain the upper'
           ' bound of the time coverage. This option is required when there is'
           ' more than one time value in the input file or when it is not'
           ' possible to provide a static value for time_coverage_start and'
           ' time_coverage_end (via input file or global overrides)',
           )

    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _var_ids = input_opts.get('variables', None)
    if _var_ids is None:
        raise MissingVariablesList()
    var_ids = [x.strip() for x in _var_ids.split(',')]

    proj_def = input_opts.get('projection', None)
    if proj_def is None:
        raise MissingProjection()

    spatial_resolution = input_opts.get('spatial_resolution', None)
    if spatial_resolution is None:
        raise MissingSpatialResolution()

    # Read options related to lat/lon coordinates
    lon_var_id = input_opts.get('lon_variable', None)
    lat_var_id = input_opts.get('lat_variable', None)
    lon_dim_id = input_opts.get('lon_dimension', lon_var_id)
    lat_dim_id = input_opts.get('lat_dimension', lat_var_id)
    _lon_reversed = input_opts.get('lon_reversed', False)
    if _lon_reversed is not False:
        lon_reversed = _lon_reversed.lower() in ('yes', 'true', '1')
    else:
        lon_reversed = _lon_reversed
    _lat_reversed = input_opts.get('lat_reversed', False)
    if _lat_reversed is not False:
        lat_reversed = _lat_reversed.lower() in ('yes', 'true', '1')
    else:
        lat_reversed = _lat_reversed

    # Read options related to y/x coordinates
    x_var_id = input_opts.get('x_variable', None)
    y_var_id = input_opts.get('y_variable', None)
    x_dim_id = input_opts.get('x_dimension', x_var_id)
    y_dim_id = input_opts.get('y_dimension', y_var_id)
    _x_reversed = input_opts.get('x_reversed', False)
    if _x_reversed is not False:
        x_reversed = _x_reversed.lower() in ('yes', 'true', '1')
    else:
        x_reversed = _x_reversed
    _y_reversed = input_opts.get('y_reversed', False)
    if _y_reversed is not False:
        y_reversed = _y_reversed.lower() in ('yes', 'true', '1')
    else:
        y_reversed = _y_reversed

    time_var_id = input_opts.get('time_variable', None)

    # Check that at least one coordinate vars couple is defined
    spatial_coords = None
    if None not in (x_var_id, y_var_id):
        spatial_coords = 'yx'
        dim1_var_id = y_var_id
        dim2_var_id = x_var_id
        dim1_id = y_dim_id
        dim2_id = x_dim_id
        dim1_reversed = y_reversed
        dim2_reversed = x_reversed
    elif None not in (lon_var_id, lat_var_id):
        spatial_coords = 'latlon'
        dim1_var_id = lat_var_id
        dim2_var_id = lon_var_id
        dim1_id = lat_dim_id
        dim2_id = lon_dim_id
        dim1_reversed = lat_reversed
        dim2_reversed = lon_reversed
    else:
        raise MissingCoordinatesSystem()

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    idf_converter.lib.apply_global_overrides(input_opts, granule)
    idf_converter.lib.apply_var_overrides(input_opts, granule)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')

    # - Time
    dtimes: numpy.typing.NDArray = numpy.array([None, ])
    if time_var_id is not None:
        _t = idf_converter.lib.extract_variable_values(f_handler, time_var_id)
        _t_mask = numpy.ma.getmaskarray(_t).copy()
        if _t_mask.all():
            raise MissingValidTime(time_var_id)

        if _t_mask.any():
            logger.warning('Some time values are masked, this is considered '
                           'bad practice. These masked values will be ignored')

            # Temporarily replace masked values by a valid one ortherwise the
            # netCDF4.num2date call will fail. The mask of the time variable
            # will be applied to the output of num2date later on.
            # At this point in the code we know that there is at least one non-
            # masked value, so we use the first we find as a placeholder.
            masked_t_ind = numpy.where(_t_mask)
            first_valid_t = _t[numpy.where(~_t_mask)[0][0]]
            _t[masked_t_ind] = first_valid_t

        _t_units = f_handler.variables[time_var_id].units
        if 'time_units' in input_opts:
            _t_units = input_opts['time_units']
        if 'time_offset' in input_opts:
            _time_offset = float(input_opts['time_offset'])
            _t = _t + _time_offset
        dtimes = netCDF4.num2date(numpy.ma.getdata(_t), _t_units)
        dtimes = idf_converter.lib.time.as_datetimes_array(dtimes)

        # Apply the same mask as the time value read from the input file to
        # ignore cells that were originally masked and were temporarily
        # replaced by valid values to avoid computation issues
        dtimes = numpy.ma.masked_where(_t_mask, dtimes, copy=False)

    # - Space
    dim1 = idf_converter.lib.extract_variable_values(f_handler, dim1_var_id,
                                                     False)
    dim2 = idf_converter.lib.extract_variable_values(f_handler, dim2_var_id,
                                                     False)

    dim1_shape = numpy.shape(dim1)
    while (2 < len(dim1_shape)) and 1 == dim1_shape[0]:
        dim1 = dim1[0]
        dim1_shape = numpy.shape(dim1)

    dim2_shape = numpy.shape(dim2)
    while (2 < len(dim2_shape)) and 1 == dim2_shape[0]:
        dim2 = dim2[0]
        dim2_shape = numpy.shape(dim2)

    """
    if 'latlon' == spatial_coords:
        # Handle longitude continuity
        dim2_delta = dim2[1:] - dim2[:-1]
        if 180.0 <= numpy.max(numpy.abs(dim2_delta)):
            lon_neg_ind = numpy.where(dim2 < 0.0)
            dim2[lon_neg_ind] = dim2[lon_neg_ind] + 360.0

        # Wrap data in [-180.0, 180[ instead of [0, 360[
        # This is not a requirement of the IDF specifications but the current
        # implementation of streamlines rendering in SEAScope will not work
        # otherwise
        dim2 = numpy.mod(dim2 + 180.0, 360.0) - 180.0
    """

    # - Depth
    depth_info = idf_converter.lib.generic_grid.get_depth_info(input_opts,
                                                               granule,
                                                               f_handler)

    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)

    dim1_size = f_handler.dimensions[dim1_id].size
    dim2_size = f_handler.dimensions[dim2_id].size

    if 'latlon' == spatial_coords:
        granule.vars['lat'] = {'array': dim1,
                               'units': 'degrees_north',
                               'datatype': dim1.dtype,
                               'options': {}}
        granule.vars['lon'] = {'array': dim2,
                               'units': 'degrees_east',
                               'datatype': dim2.dtype,
                               'options': {}}
    else:
        granule.vars['y'] = {'array': dim1,
                             'datatype': dim1.dtype,
                             'options': {}}
        granule.vars['x'] = {'array': dim2,
                             'datatype': dim2.dtype,
                             'options': {}}
    granule.dims['y'] = dim1_size
    granule.dims['x'] = dim2_size

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_resolution
    granule.meta['idf_spatial_resolution_units'] = 'm'

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)
    granule.meta['idf_granule_id'] = granule_name

    dim1_info = (dim1_id, dim1_reversed, None, None)
    dim2_info = (dim2_id, dim2_reversed, None, None)

    if 'indicative_times' in input_opts:
        dtimes = input_opts['indicative_times']

    generic_processor = idf_converter.lib.generic_grid.single_granule
    if 1 < len(dtimes):
        generic_processor = idf_converter.lib.generic_grid.multi_granule
    for extracted_data in generic_processor(granule, f_handler, var_ids,
                                            depth_info, dim1_info, dim2_info,
                                            input_opts, output_opts,
                                            time_var_id, dtimes):
        yield extracted_data

    f_handler.close()



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-07
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import datetime
import scipy.signal
import numpy.typing
import idf_converter.lib
import idf_converter.lib.geo
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult
from idf_converter.lib.types import TransformsList, Granule

logger = logging.getLogger(__name__)

DATA_MODEL = idf_converter.readers.netcdf_grid_latlon.DATA_MODEL
WindConfig = typing.Dict[str, typing.Dict[str, str]]
WindVariables = typing.Tuple[typing.List[str], WindConfig]
MutableIndexSlices = typing.List[typing.Union[int, slice]]
IndexSlices = typing.Sequence[typing.Union[int, slice]]
SPEED_VARS = ('wind', 'wind_speed_LF', 'wind_speed_MF', 'wind_speed_AW',
              'wind_speed_TC')


class InputPathMissing(Exception):
    """Error raised when the "path" input option has not been specified by the
    user."""
    pass


class TimeUnitsNotSupported(Exception):
    """Error raised when the units of the time variable are not support by this
    reader."""
    def __init__(self, time_units: str) -> None:
        self.time_units = time_units


class InvalidWindVariable(ValueError):
    """Error raised when the value passed to the wind_variables option does not
    match the expected pattern."""
    def __init__(self, value: str) -> None:
        self.wind_variables_str = value


class NoValidMinute(Exception):
    """Error raised when splitting into swaths is impossible due to the lack
    of valid values in the minutes array."""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('    split-swaths\tSet to yes|true|1 if the converter should create'
           ' one output file for each swath (simulating a gridded L2) instead'
           ' of the original data structure (one file containing all swaths)',
           )
    return ('\n'.join(inp), '\t'.join(out))


def get_wind_transforms(output_opts: OutputOptions,
                        wind_vars: WindConfig,
                        original_var_ids: typing.List[str],
                        has_asc_desc: bool,
                        ) -> TransformsList:
    """"""
    # Add transform for converting (module, dir) to (eastward, northward)
    transforms: TransformsList = []

    to_remove = []
    vec_cfg = {}
    for wind_var_id, wind_var_cfg in wind_vars.items():
        speed_var_id = wind_var_cfg['speed']
        dir_var_id = wind_var_cfg['direction']

        if has_asc_desc:
            speed_asc_id = f'{speed_var_id}_asc'
            if speed_asc_id not in original_var_ids:
                to_remove.append(speed_asc_id)

            speed_desc_id = f'{speed_var_id}_desc'
            if speed_desc_id not in original_var_ids:
                to_remove.append(speed_desc_id)

            dir_asc_id = f'{dir_var_id}_asc'
            if dir_asc_id not in original_var_ids:
                to_remove.append(dir_asc_id)

            dir_desc_id = f'{dir_var_id}_desc'
            if dir_desc_id not in original_var_ids:
                to_remove.append(dir_desc_id)

            vec_cfg[f'{wind_var_id}_asc'] = {'direction': dir_asc_id,
                                             'module': speed_asc_id,
                                             'radians': False,
                                             'angle_to_east': numpy.pi / 2.0,
                                             'clockwise': True,
                                             'meteo': False}
            vec_cfg[f'{wind_var_id}_desc'] = {'direction': dir_desc_id,
                                              'module': speed_desc_id,
                                              'radians': False,
                                              'angle_to_east': numpy.pi / 2.0,
                                              'clockwise': True,
                                              'meteo': False}
        else:
            if speed_var_id not in original_var_ids:
                to_remove.append(speed_var_id)

            if dir_var_id not in original_var_ids:
                to_remove.append(dir_var_id)

            vec_cfg[wind_var_id] = {'direction': dir_var_id,
                                    'module': speed_var_id,
                                    'radians': False,
                                    'angle_to_east': numpy.pi / 2.0,
                                    'clockwise': True,
                                    'meteo': False}

    wind_targets = list(vec_cfg.keys())
    to_remove = list(set(to_remove))

    if 0 < len(wind_targets):
        # Add eastward_ and northward_ variables to the export targets
        targets = output_opts['__export']

        for wind_var_id in wind_targets:
            targets.append(f'eastward_{wind_var_id}')
            targets.append(f'northward_{wind_var_id}')

        transforms.append(('dir2vectors', {'targets': wind_targets,
                                           'configs': vec_cfg}))

    if 0 < len(to_remove):
        # If eastward/northward variables carry the wind speed and direction
        # information so there is no need to keep it in the output
        transforms.append(('remove_vars', {'targets': to_remove}))

    return transforms


def create_granule(input_opts: InputOptions,
                   output_opts: OutputOptions,
                   pass_suffix: str,
                   start_dt: datetime.datetime,
                   stop_dt: datetime.datetime,
                   spatial_resolution: float,
                   lat: numpy.typing.NDArray,
                   lon: numpy.typing.NDArray,
                   values: typing.Dict[str, numpy.typing.NDArray],
                   units: typing.Dict[str, str],
                   ) -> Granule:
    """"""
    # Format granule
    granule_name, _ = os.path.splitext(os.path.basename(input_opts['path']))

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    _swath_granule_name = f'{granule_name}_{start_dt:%H%M}_{stop_dt:%H%M}'
    _granule_name = f'{_swath_granule_name}{pass_suffix}'
    granule.meta['idf_granule_id'] = _granule_name

    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_resolution
    granule.meta['idf_spatial_resolution_units'] = 'm'

    for var_id, var_values in values.items():
        var_units = units.get(var_id, '')
        granule.vars[var_id] = {'array': var_values,
                                'name': var_id,
                                'units': var_units,
                                'valid_min': numpy.nanmin(var_values),
                                'valid_max': numpy.nanmax(var_values),
                                'datatype': var_values.dtype,
                                'options': {}}

    _lon, trunc_to_0 = idf_converter.lib.mod360(lon - lon[0], 9)
    _lon = _lon + lon[0]

    if trunc_to_0 is True:
        logger.warning('Some longitude values very close to 0 degrees have '
                       'been truncated to 0 to avoid underflow errors')

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': _lon,
                           'units': 'degrees_east',
                           'datatype': _lon.dtype,
                           'options': {}}
    granule.dims['lat'] = lat.size
    granule.dims['lon'] = lon.size

    return granule


def get_swath_minute_bounds(minute: numpy.typing.ArrayLike,
                            min_swath_duration: int = 80
                            ) -> typing.Tuple[numpy.typing.NDArray,
                                              numpy.typing.NDArray]:
    """"""
    valid_minute = minute.compressed()
    if valid_minute.size == 0:
        raise NoValidMinute()
    hist, _ = numpy.histogram(valid_minute, bins=1441, range=(0, 1441))

    valid_index = numpy.where(hist > 0)[0]
    valid_index.sort()

    # Compute mode for minute differences in order to identify actual gaps
    # later on
    dminute, count = numpy.unique(valid_index[1:] - valid_index[:-1],
                                  return_counts=True)
    nominal_dminute = dminute[count.argmax()]

    # Smoothen the histogram to remove little gaps
    span = 10
    smooth_hist = numpy.convolve(hist,
                                 numpy.ones(span * 2 + 1) / (span * 2 + 1),
                                 mode="same")

    # Find indexes of local minima on the smooth histogram
    smooth_min_index, _ = scipy.signal.find_peaks(-1 * smooth_hist,
                                                  distance=min_swath_duration)

    # Compute local minima on the actual histogram
    _minute0 = [valid_index[0]]
    _minute1 = []
    for i in range(0, smooth_min_index.size):
        smooth_index = smooth_min_index[i]

        # The index of the local min in the smoothed histogram might be a bin
        # with a 0 count in the original histogram so we need to get the index
        # of the closest non empty bin
        closest_data_index = numpy.searchsorted(valid_index, smooth_index)
        hist_index = valid_index[closest_data_index]

        prev_mask = (valid_index < hist_index)
        valid_prev = numpy.array([])
        if prev_mask.any():
            valid_prev = valid_index[numpy.where(prev_mask)]
        next_mask = (valid_index > hist_index)
        valid_next = numpy.array([])
        if next_mask.any():
            valid_next = valid_index[numpy.where(next_mask)]

        # Make sure local min in the smoothed histogram is also a local min in
        # the original histogram
        prev_index = hist_index
        if 0 < valid_prev.size:
            prev_index = valid_prev[-1]

        next_index = hist_index
        if 0 < valid_next.size:
            next_index = valid_next[0]

        prev_dist = hist_index - prev_index
        next_dist = next_index - hist_index
        if 2 * nominal_dminute < prev_dist + next_dist:
            hist_index = prev_index
        else:
            closer_to_prev = (prev_dist < next_dist)
            closer_to_next = (prev_dist > next_dist)
            equidistant = (prev_dist == next_dist)
            prev_value = hist[prev_index]
            next_value = hist[next_index]
            if closer_to_prev or (equidistant and (prev_value < next_value)):
                # Check left
                _smaller_mask = (hist[valid_prev] < hist[hist_index])
                smaller_mask = numpy.cumprod(_smaller_mask[::-1])[::-1]
                if (0 < smaller_mask).any():
                    hist_index = valid_prev[numpy.where(smaller_mask)].min()
            elif closer_to_next or (equidistant and (next_value < prev_value)):
                # Check right
                _smaller_mask = (hist[valid_next] < hist[hist_index])
                smaller_mask = numpy.cumprod(_smaller_mask)
                if (0 < smaller_mask).any():
                    hist_index = valid_next[numpy.where(smaller_mask)].max()

        _minute1.append(hist_index)

        next_mask = (valid_index > hist_index)
        if next_mask.any():
            m0_index = valid_index[numpy.where(next_mask)].min()
            _minute0.append(m0_index)

    minute0 = numpy.array(_minute0)
    if len(_minute0) > len(_minute1):
        minute1 = numpy.concatenate([_minute1, [valid_index[-1]]])
    else:
        minute1 = numpy.array(_minute1)

    return minute0, minute1


def split_swaths(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 original_var_ids: typing.List[str],
                 day: datetime.datetime,
                 minute: numpy.typing.ArrayLike,
                 lat: numpy.typing.NDArray,
                 lon: numpy.typing.NDArray,
                 pass_suffix: str,
                 values: typing.Dict[str, numpy.typing.NDArray],
                 units: typing.Dict[str, str],
                 wind_vars: WindConfig
                 ) -> typing.Iterator[ReaderResult]:
    """"""
    nlon = lon.size

    minute0, minute1 = get_swath_minute_bounds(minute)

    # Extract mask for the first speed variable found in SPEED_VARS
    wind_mask = numpy.full(minute.shape, False, dtype='bool')
    for var_id in SPEED_VARS:
        if var_id in values:
            wind_mask = numpy.ma.getmaskarray(values[var_id])
            break

    dlat = lat[1:] - lat[:-1]
    dlon = lon[1:] - lon[:-1]
    spatial_resolution = min([abs(dlat.mean()), abs(dlon.mean())]) * 111000.

    concat: typing.Callable[[typing.Sequence[numpy.typing.ArrayLike],
                             typing.Optional[int]],
                            numpy.typing.ArrayLike]
    concat = numpy.ma.concatenate
    for _minute0, _minute1 in zip(minute0, minute1):
        indpass = numpy.where((minute >= _minute0) &
                              (minute <= _minute1) &
                              (~wind_mask))

        if indpass[0].size == 0:
            continue
        if indpass[0].min() == indpass[0].max():
            logger.warning('Skipping swath containing a single latitude')
            continue

        lat_slice = slice(indpass[0].min(), indpass[0].max() + 1)
        swath_lat = lat[lat_slice]

        indlon = numpy.unique(indpass[1])
        if indlon.size == 1:
            dindlon = numpy.array([0])
        else:
            dindlon = indlon[1:] - indlon[:-1]

        swath_values = {}
        if (indlon[-1] - indlon[0]) <= (nlon - dindlon.max()):
            lon_slice = slice(indlon[0], indlon[-1] + 1)
            swath_lon = lon[lon_slice]
            swath_minute = minute[lat_slice, lon_slice].copy()
            for var_id, var_values in values.items():
                swath_values[var_id] = var_values[lat_slice, lon_slice].copy()
        else:
            indsplit = dindlon.argmax()
            lon_slice0 = slice(indlon[indsplit + 1], nlon)
            lon_slice1 = slice(0, indlon[indsplit] + 1)
            swath_lon = concat((lon[lon_slice0], lon[lon_slice1]), axis=0)
            swath_minute = concat((minute[lat_slice, lon_slice0],
                                   minute[lat_slice, lon_slice1]),
                                  axis=1)
            for var_id, var_values in values.items():
                slice0_values = var_values[lat_slice, lon_slice0]
                slice1_values = var_values[lat_slice, lon_slice1]
                swath_values[var_id] = concat((slice0_values, slice1_values),
                                              axis=1)

        before_ind = numpy.where(swath_minute < _minute0)
        after_ind = numpy.where(swath_minute > _minute1)
        for var_id, var_values in swath_values.items():
            var_values[before_ind] = numpy.ma.masked
            var_values[after_ind] = numpy.ma.masked

        start_dt = day + datetime.timedelta(seconds=int(_minute0 * 60))
        stop_dt = day + datetime.timedelta(seconds=int(_minute1 * 60))
        granule = create_granule(input_opts, output_opts, pass_suffix,
                                 start_dt, stop_dt, spatial_resolution,
                                 swath_lat, swath_lon, swath_values, units)

        output_opts['__export'] = list(swath_values.keys())

        # Add transforms
        transforms = get_wind_transforms(output_opts, wind_vars,
                                         original_var_ids, False)

        yield input_opts, output_opts, granule, transforms


def parse_wind_vars(input_opts: InputOptions
                    ) -> WindVariables:
    """"""
    required_vars: typing.List[str] = []
    wind_vars: WindConfig = {}
    _wind_vars = input_opts.get('wind_variables', None)
    if _wind_vars is None:
        return (required_vars, wind_vars)

    for wind_cfg_str in _wind_vars.split(','):
        if 0 >= len(wind_cfg_str.strip()):
            continue

        wind_cfg_elements = wind_cfg_str.split(':')
        if 3 == len(wind_cfg_elements):
            output_var_id, speed_var_id, direction_var_id = wind_cfg_elements
            wind_vars[output_var_id] = {'speed': speed_var_id,
                                        'direction': direction_var_id}
            required_vars.append(speed_var_id)
            required_vars.append(direction_var_id)
        else:
            raise InvalidWindVariable(wind_cfg_str)

    return (list(set(required_vars)), wind_vars)


def compute_slices(variable: netCDF4.Variable,
                   pass_dim_id: str
                   ) -> typing.Tuple[IndexSlices, IndexSlices]:
    """Compute data slices assuming all variables have the same dimensions"""
    asc_slices: MutableIndexSlices = []
    desc_slices: MutableIndexSlices = []
    for dim_id in variable.dimensions:
        if dim_id == pass_dim_id:
            asc_slices.append(0)
            desc_slices.append(1)
        else:
            asc_slices.append(slice(None, None, None))
            desc_slices.append(slice(None, None, None))

    return asc_slices, desc_slices


def split_ascending_descending(input_opts: InputOptions,
                               output_opts: OutputOptions,
                               original_var_ids: typing.List[str],
                               day: datetime.datetime,
                               minutes: numpy.typing.NDArray,
                               lon_var_id: str,
                               lat_var_id: str,
                               values: typing.Dict[str, numpy.typing.NDArray],
                               units: typing.Dict[str, str],
                               wind_vars: WindConfig,
                               asc_slices: IndexSlices,
                               desc_slices: IndexSlices
                               ) -> typing.Iterator[ReaderResult]:
    """"""
    # Format extracted data into input options before passing them to the
    # generic netCDF latlon grid reader
    input_opts['lon_variable'] = lon_var_id
    input_opts['lat_variable'] = lat_var_id

    var_ids_by_pass = []
    input_opts['extra_values'] = {}
    for var_id in values.keys():
        var_units = units.get(var_id, '')

        asc_values = values[var_id][asc_slices]
        input_opts['extra_values'][f'{var_id}_asc'] = {'array': asc_values,
                                                       'units': var_units,
                                                       'dims': ['lat', 'lon']}

        desc_values = values[var_id][desc_slices]
        input_opts['extra_values'][f'{var_id}_desc'] = {'array': desc_values,
                                                        'units': var_units,
                                                        'dims': ['lat', 'lon']}
        var_ids_by_pass.append(f'{var_id}_asc')
        var_ids_by_pass.append(f'{var_id}_desc')

    input_opts['variables'] = ','.join(var_ids_by_pass)

    # Time coverage
    start_minutes = int(minutes.min())
    start_dt = day + datetime.timedelta(seconds=start_minutes * 60)
    start_str = f'time_coverage_start:"{start_dt:%Y-%m-%dT%H:%M:%SZ}"'

    stop_minutes = int(minutes.max())
    stop_dt = day + datetime.timedelta(seconds=stop_minutes * 60)
    stop_str = f'time_coverage_end:"{stop_dt:%Y-%m-%dT%H:%M:%SZ}"'

    input_opts['global_overrides'] = ','.join([start_str, stop_str])
    del input_opts['time_variable']  # Global attributes provide time coverage

    generic_latlon_grid = idf_converter.readers.netcdf_grid_latlon
    generic = generic_latlon_grid.read_data(input_opts, output_opts)
    input_opts, output_opts, granule, transforms = next(generic)

    # Add transform for converting (module, dir) to (eastward, northward)
    to_remove = []
    vec_cfg = {}
    for wind_var_id, wind_var_cfg in wind_vars.items():
        speed_var_id = wind_var_cfg['speed']
        dir_var_id = wind_var_cfg['direction']

        to_remove.append(f'{speed_var_id}_asc')
        to_remove.append(f'{speed_var_id}_desc')
        to_remove.append(f'{dir_var_id}_asc')
        to_remove.append(f'{dir_var_id}_desc')

        vec_cfg[f'{wind_var_id}_asc'] = {'direction': f'{dir_var_id}_asc',
                                         'module': f'{speed_var_id}_asc',
                                         'radians': False,
                                         'angle_to_east': numpy.pi / 2.0,
                                         'clockwise': True,
                                         'meteo': False}
        vec_cfg[f'{wind_var_id}_desc'] = {'direction': f'{dir_var_id}_desc',
                                          'module': f'{speed_var_id}_desc',
                                          'radians': False,
                                          'angle_to_east': numpy.pi / 2.0,
                                          'clockwise': True,
                                          'meteo': False}
    wind_targets = list(vec_cfg.keys())

    if 0 < len(wind_targets):
        # Add eastward_ and northward_ variables to the export targets
        targets = output_opts['__export']

        for wind_var_id in wind_targets:
            targets.append(f'eastward_{wind_var_id}')
            targets.append(f'northward_{wind_var_id}')

        transforms.append(('dir2vectors', {'targets': wind_targets,
                                           'configs': vec_cfg}))

        # If eastward/northward variables carry the wind speed and direction
        # information so there is no need to keep it in the output
        transforms.append(('remove_vars', {'targets': to_remove}))

    yield input_opts, output_opts, granule, transforms


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    pass_dim_id = input_opts.get('pass_dimension', 'node')
    lat_var_id = input_opts.get('lat_variable', 'lat')
    lon_var_id = input_opts.get('lon_variable', 'lon')
    time_var_id = input_opts.get('time_variable', 'minute')

    variables = input_opts.get('variables', 'wind')
    original_var_ids = [_.strip() for _ in variables.split(',')
                        if 0 < len(_.strip())]

    required_var_ids, wind_vars = parse_wind_vars(input_opts)

    var_ids = original_var_ids + required_var_ids
    var_ids = list(set(var_ids))
    if 0 >= len(var_ids):
        logger.warning('No variable to extract')
        raise idf_converter.lib.EarlyExit()

    # Read variables
    f_handler = netCDF4.Dataset(input_path, 'r')

    lon = f_handler.variables[lon_var_id][:]
    lat = f_handler.variables[lat_var_id][:]
    time_units = f_handler.variables[time_var_id].units
    time_values = f_handler.variables[time_var_id][:]

    values = {}
    units = {}
    for var_id in var_ids:
        # The valid_min and valid_max attributes are incorrect, causing masking
        # issues. Here we disable netCDF auto-masking for this variable as a
        # workaround.
        f_handler.variables[var_id].set_auto_mask(False)
        _values = f_handler.variables[var_id][:]
        units[var_id] = f_handler.variables[var_id].units
        if var_id in SPEED_VARS:
            # Mask negative values for wind speed
            values[var_id] = numpy.ma.masked_less(_values, 0)
        else:
            values[var_id] = _values

    # Get temporal coverage
    day_attr_name = 'day_of_month_of observation'
    if hasattr(f_handler, day_attr_name):
        day = datetime.datetime(f_handler.year_of_observation,
                                f_handler.month_of_observation,
                                f_handler.getncattr(day_attr_name))
    else:
        day = datetime.datetime(f_handler.year_of_observation,
                                f_handler.month_of_observation,
                                f_handler.day_of_month_of_observation)

    # Compute slices for ascending and descending passes assuming all variables
    # have the same dimensions
    _asc_slices, _desc_slices = compute_slices(f_handler.variables[var_ids[0]],
                                               pass_dim_id)
    f_handler.close()

    asc_slices = tuple(_asc_slices)
    desc_slices = tuple(_desc_slices)

    # Convert time to minutes
    if 'minute' in time_units:
        minutes = time_values
    elif 'hour' in time_units:
        minutes = time_values * 60
    else:
        raise TimeUnitsNotSupported(time_units)

    use_swaths_splitting = output_opts.get('split-swaths', 'n')
    if use_swaths_splitting.lower() in ('1', 'y', 'yes'):
        pass_minutes = minutes[asc_slices]
        asc_values = {}
        for var_id, var_values in values.items():
            asc_values[var_id] = var_values[asc_slices]
        yield from split_swaths(input_opts, output_opts, original_var_ids,
                                day, pass_minutes, lat, lon, '_asc',
                                asc_values, units, wind_vars)

        pass_minutes = minutes[desc_slices]
        desc_values = {}
        for var_id, var_values in values.items():
            desc_values[var_id] = var_values[desc_slices]
        yield from split_swaths(input_opts, output_opts, original_var_ids,
                                day, pass_minutes, lat, lon, '_desc',
                                desc_values, units, wind_vars)
    else:
        yield from split_ascending_descending(input_opts, output_opts,
                                              original_var_ids, day, minutes,
                                              lon_var_id, lat_var_id, values,
                                              units, wind_vars, asc_slices,
                                              desc_slices)



# vim: ts=4:sts=4:sw=4
#
# @author <sylvain.herledan@oceandatalab.com>
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides methods to handle parameters passed on the command line
and to perform the conversion of the input file to IDF format.
"""

import os
import sys
import numpy
import typing
import logging
import argparse
import warnings
import idf_converter
import idf_converter.lib
import idf_converter.lib.types
import idf_converter.writer
from pkg_resources import iter_entry_points

logger = logging.getLogger(__name__)
MainArgsResult = typing.Tuple[argparse.Namespace, typing.Any]
ComplexArgsResult = typing.Dict[str, str]
ArgsResult = typing.Dict[str, typing.Any]
PostProcessingOption = typing.Tuple[str, typing.Dict[str, typing.Any]]


class InvalidOptionsPath(ValueError):
    """Error raised when options must be read from a file that does not exist.
    """
    pass


class MalformattedInputOption(ValueError):
    """Error raised when the user provided an input option which is not in
    the key=value format."""

    def __init__(self, bad_option: str) -> None:
        self.bad_option = bad_option
        super().__init__()


class MalformattedOutputOption(ValueError):
    """Error raised when the user provided an output option which is not in
    the key=value format."""

    def __init__(self, bad_option: str) -> None:
        self.bad_option = bad_option
        super().__init__()


class FormatterMissing(ValueError):
    """Error raised when a reader requested a formatting operation that is not
    supported."""

    def __init__(self, fmt_name: str) -> None:
        self.fmt_name = fmt_name
        super(FormatterMissing, self).__init__()


class PostProcessorMissing(ValueError):
    """Error raised when output options mention a post-processing but the
    associated processor is unknown / not supported."""

    def __init__(self, processor_name: str) -> None:
        self.processor_name = processor_name
        super(PostProcessorMissing, self).__init__()


class HelpRequested(Exception):
    pass


class VersionRequested(Exception):
    pass


class NoInputOptions(Exception):
    pass


class NoOutputOptions(Exception):
    pass


class NoReader(Exception):
    pass


def _get_workflow_help(reader_mod: typing.Any, idf_version: str
                       ) -> typing.Tuple[str, str]:
    """"""
    reader_input_help, reader_output_help = reader_mod.help()
    fmt = idf_converter.lib.get_idf_formatter(idf_version)
    fmt_input_help, fmt_output_help = fmt.help()
    data_model_mod = fmt.get_data_model(reader_mod.DATA_MODEL)
    idf_input_help, idf_output_help = data_model_mod.help()
    writer_input_help, writer_output_help = idf_converter.writer.help()

    input_help = f'{reader_input_help}\n{idf_input_help}\n{fmt_input_help}'
    output_help = f'{reader_output_help}\n{idf_output_help}\n{fmt_output_help}'

    input_help = f'{input_help}\n{writer_input_help}'
    output_help = f'{output_help}\n{writer_output_help}'

    # Remove empty lines
    input_help = '\n'.join([_ for _ in input_help.split('\n')
                            if 0 < len(_.strip())])
    output_help = '\n'.join([_ for _ in output_help.split('\n')
                             if 0 < len(_.strip())])

    return (input_help, output_help)


def show_help(input_readers: typing.Dict[str, typing.Any],
              main_args: argparse.Namespace) -> None:
    """
    Display help message explaining how to use the conversion command.

    Parameters
    ----------
    input_readers: dict
        TODO
    main_args: argparse.Namespace
        TODO
    """
    reader = main_args.t if main_args.t is not None else 'READER'
    print('\nUsage:\n'
          f'idf-converter -t {reader} -i key=value [key=value ...] '
          '-o key=value [key=value ...]\n[--idf-version=_version_] '
          '[--verbose] [--debug]')

    if 'READER' == reader:
        readers = '\n'.join(f'    {x}' for x in list(input_readers.keys()))
        print('\nReader\n'
              '  -t READER\tIdentifier of the reader which will extract data '
              'from the input file(s)\n\n'
              f'  Supported readers:\n{readers}\n')

        print('\nInput parameters\n'
              '  -i key=value [key=value ...]\n'
              '  -i FILE_PATH@\twhere FILE_PATH points to a text file which '
              'contains key=value pairs\n\n'
              '  Supported keys depend on the reader you choose, you can get '
              'the list\n'
              '  of the input parameters supported by a reader by using:\n\n'
              '    idf-converter -t READER -h\n')
        print('\nOutput parameters\n'
              '  -o key=value [key=value ...]\n'
              '  -o FILE_PATH@\twhere FILE_PATH points to a text file which '
              'contains key=value pairs\n\n'
              '  Supported keys depend on the reader you choose, you can get '
              'the list\n'
              '  of the output parameters supported by a reader by using:\n\n'
              '    idf-converter -t READER -h\n')
    else:
        input_help, output_help = _get_workflow_help(input_readers[reader],
                                                     main_args.idf_version)
        print('\nInput parameters\n'
              '  -i key=value [key=value ...]\n'
              '  -i FILE_PATH@\twhere FILE_PATH points to a text file which '
              'contains key=value pairs\n\n'
              f'  Supported keys are:\n{input_help}\n')

        print('\nOutput parameters\n'
              '  -o key=value [key=value ...]\n'
              '  -o FILE_PATH@\twhere FILE_PATH points to a text file which '
              'contains key=value pairs\n\n'
              f'  Supported keys are:\n{output_help}\n')

    print('\nOptions\n'
          '  -h, --help\tShow this help message and exit\n'
          '  --idf-version\tVersion of the IDF specifications that the '
          'converter will use to produce output files.\n'
          '  --verbose\tShow informational messages.\n'
          '  --debug\tShow debug messages.\n'
          '  --version\tPrint the version of the IDF converter and exit\n'
          '  --paranoid\tTreat all warnings as errors\n'
          '  --treat-numpy-errors-as-warnings\tNumerical issues detected by '
          'numpy will only display a warning message instead of raising an '
          'exception. This option is meant for debugging purpose and all the '
          'readers must be able to run without this option in production\n\n')


def parse_main_args(readers_names: typing.List[str],
                    inputs: typing.Optional[typing.List[str]] = None
                    ) -> MainArgsResult:
    """Parse generic arguments, common to all readers.

    Parameters
    ----------
    readers_names: list
        TODO

    Returns
    -------
    tuple
        TODO
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-t', nargs='?', choices=readers_names)
    parser.add_argument('-o', nargs='+', action='append')
    parser.add_argument('-h', '--help', action='store_true', default=False)
    parser.add_argument('--idf-version', choices=('1.0',), default='1.0')
    parser.add_argument('--version', action='store_true', default=False)
    parser.add_argument('--verbose', action='store_true', default=False)
    parser.add_argument('--debug', action='store_true', default=False)
    parser.add_argument('--paranoid', action='store_true', default=False)
    parser.add_argument('--treat-numpy-errors-as-warnings',
                        action='store_true', default=False)

    args, unknown = parser.parse_known_args(inputs)
    return args, unknown


def parse_reader_args(inputs: typing.List[str]
                      ) -> typing.Tuple[argparse.Namespace, typing.List[str]]:
    """Parse arguments related to input file, i.e arguments relevant for the
    data readers.

    Parameters
    ----------
    inputs: list
        TODO

    Returns
    -------
    argparse.Namespace
        TODO
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-i', nargs='+', action='append')
    args, unknown = parser.parse_known_args(inputs)
    return args, unknown


def parse_complex_list_argument(inputs: typing.Optional[typing.List[str]],
                                format_exc: typing.Any) -> ComplexArgsResult:
    """Parse command line arguments expected either as a list of key=value
    tokens, a path prefixed by the '@' symbol leading to a file that contains
    these tokens or both.

    Parameters
    ----------
    inputs: list
        TODO
    format_exc: type
        TODO

    Returns
    -------
    dict
        TODO
    """
    if inputs is None:
        return {}

    _options = [x for tokens in inputs for x in tokens]
    options_files = []
    options = []
    for _option in _options:
        if _option.endswith('@'):
            options_files.append(_option[:-1])
        else:
            options.append(_option)

    # Load options from files
    for relative_path in options_files:
        full_path = os.path.abspath(relative_path)
        if not os.path.exists(full_path):
            raise FileNotFoundError(full_path)

        with open(full_path, 'rt') as f:
            raw_text = f.read()
        file_option = raw_text.strip()
        if 0 < len(file_option):
            # Prepend option from file so that it can be overwritten by
            # the command line options
            options.insert(0, file_option)

    result = {}
    parse_opt = idf_converter.lib.parse_opt
    for opt_key, opt_value in parse_opt(' '.join(options), ' ', '='):
        result[opt_key] = opt_value

    return result


def parse_args(input_readers: typing.Dict[str, typing.Any]) -> ArgsResult:
    """Parse arguments passed on the command.

    Parameters
    ----------
    input_readers: dict
        TODO

    Returns
    -------
    dict
        TODO
    """
    _parse_complex_args = parse_complex_list_argument
    readers_names = list(input_readers.keys())

    args, extra = parse_main_args(readers_names)
    if args.help:
        show_help(input_readers, args)
        raise HelpRequested()

    if args.version:
        raise VersionRequested()

    output_options: ComplexArgsResult = {}
    if args.o is not None:
        output_options = _parse_complex_args(args.o, MalformattedOutputOption)

    reader_args, extra = parse_reader_args(extra)

    input_options: ComplexArgsResult = {}
    if reader_args.i is not None:
        input_options = _parse_complex_args(reader_args.i,
                                            MalformattedInputOption)

    # Only one extra parameter is allowed: it has to be the path of a text file
    # which contains preset options
    extra_args = None
    extra_input_options = None
    extra_output_options = None
    if 0 < len(extra):
        if not extra[0].endswith('@'):
            raise InvalidOptionsPath(extra[0])
        with open(extra[0][:-1], 'rt') as f:
            raw_text = f.read()
        extra_opts = [_.strip() for _ in raw_text.split()]
        extra_args, extra_extra = parse_main_args(readers_names, extra_opts)
        extra_reader_args, _ = parse_reader_args(extra_extra)
        extra_input_options = _parse_complex_args(extra_reader_args.i,
                                                  MalformattedInputOption)
        extra_output_options = _parse_complex_args(extra_args.o,
                                                   MalformattedOutputOption)
        # Include extra options only if they have not been defined directly
        # on the command line
        for key in extra_input_options.keys():
            if key not in input_options.keys():
                input_options[key] = extra_input_options[key]

        for key in extra_output_options.keys():
            if key not in output_options.keys():
                output_options[key] = extra_output_options[key]

    reader_type = None
    if args.t is not None:
        reader_type = args.t
    elif extra_args is not None and extra_args.t is not None:
        reader_type = extra_args.t

    if reader_type is None:
        raise NoReader()

    # Process output options
    if 0 >= len(output_options.keys()):
        raise NoOutputOptions()

    # Process input options
    if 0 >= len(input_options.keys()):
        raise NoInputOptions()

    output_options['idf_version'] = args.idf_version

    result = {'reader': reader_type,
              'output_options': output_options,
              'input_options': input_options,
              'verbose': args.verbose,
              'debug': args.debug,
              'paranoid': args.paranoid,
              'numpy_errors_are_warnings': args.treat_numpy_errors_as_warnings}

    return result


def parse_post_processor(output_options: idf_converter.lib.types.OutputOptions
                         ) -> typing.Optional[PostProcessingOption]:
    """"""
    if 'postprocessor' not in output_options:
        return None

    post_processor_cfg = output_options['postprocessor']
    result = idf_converter.lib.parse_transform_option(post_processor_cfg)
    return result


def idf_converter_script() -> None:
    """Implementation of the idf-converter command."""
    main_logger = logging.getLogger()
    handler = logging.StreamHandler()
    handler.setLevel(logging.DEBUG)
    main_logger.addHandler(handler)
    main_logger.setLevel(logging.WARN)

    # Retrieve input options help from readers
    input_readers = {}
    for reader in iter_entry_points('idf.input.readers'):
        reader_impl = reader.load()
        input_readers[reader.name] = reader_impl

    try:
        args = parse_args(input_readers)
    except (HelpRequested, VersionRequested):
        # Not an error
        sys.exit(0)
    except NoReader:
        logger.error('It is mandatory to select a reader with -t.')
        sys.exit(1)
    except NoOutputOptions:
        logger.error('It is mandatory to provide output options with -o.')
        sys.exit(1)
    except NoInputOptions:
        logger.error('It is mandatory to provide input options with -i.')
        sys.exit(1)
    except MalformattedInputOption as e:
        logger.debug(str(e))
        logger.error('Input options must follow the "key=value" format.'
                     f'The "{e.bad_option}" option you provided does not.')
        sys.exit(1)
    except MalformattedOutputOption as e:
        logger.debug(str(e))
        logger.error('Output options must follow the "key=value" format.'
                     f'The "{e.bad_option}" option you provided does not.')
        sys.exit(1)

    if args['verbose']:
        main_logger.setLevel(logging.INFO)
    if args['debug']:
        main_logger.setLevel(logging.DEBUG)

    if args['paranoid']:
        # Useful to detect the origin of FutureWarning
        logger.warning('All warnings will be treated as errors')
        warnings.filterwarnings('error')

    # Raise exceptions for all floating point issues detected by numpy
    numpy.seterr(all='raise')
    if args['numpy_errors_are_warnings'] is True:
        logger.warning('Numpy floating point errors will be treated as '
                       'warnings. Please note that this option is only meant '
                       'for debugging purpose and should not be used in '
                       'production.')
        numpy.seterr(all='warn')

    reader_impl = input_readers[args['reader']]
    result = reader_impl.read_data(args['input_options'],
                                   args['output_options'])

    _post_processors = iter_entry_points('idf.postprocessors')
    post_processors = {_.name: _ for _ in _post_processors}

    try:
        for input_options, output_options, granule, formatter_jobs in result:
            logger.debug('\tRead: done')

            # Add virtual variables
            idf_converter.lib.add_virtual_variables(output_options,
                                                    formatter_jobs)

            # Remove variables that should not be saved in the IDF file
            idf_converter.lib.remove_variables(output_options, formatter_jobs)

            # Add data storage jobs
            fmt = idf_converter.lib.get_idf_formatter(granule.idf_version)
            fmt.apply_data_storage_policy(input_options, output_options,
                                          granule, formatter_jobs)

            _fmts = iter_entry_points('idf.formatters')
            formatters = {fmt.name: fmt for fmt in _fmts}
            for fmt_name, fmt_args in formatter_jobs:
                if fmt_name not in formatters:
                    logger.debug(f'Missing formatter: {fmt_name}')
                    raise FormatterMissing(fmt_name)

                fmt_method = formatters[fmt_name].load()
                result = fmt_method(input_options, output_options, granule,
                                    **fmt_args)
                input_options, output_options, granule = result
            logger.debug('\tFormat: done')

            post_processor_option = parse_post_processor(output_options)
            if post_processor_option is None:
                idf_converter.writer.as_idf(input_options, output_options,
                                            granule)
            else:
                processor_name, processor_args = post_processor_option
                if processor_name not in post_processors:
                    raise PostProcessorMissing(processor_name)

                processor = post_processors[processor_name].load()
                processor_result = processor(input_options, output_options,
                                             granule, **processor_args)
                for _ in processor_result:
                    sub_input_options, sub_output_options, sub_granule = _
                    idf_converter.writer.as_idf(sub_input_options,
                                                sub_output_options,
                                                sub_granule)
            logger.debug('\tWrite: done')
    except idf_converter.lib.EarlyExit:
        sys.exit(0)

    sys.exit(0)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions
from idf_converter.lib.types import ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'time_dependent'


class InputPathMissing(ValueError):
    """Error raised when the input options have no "path" entry"""
    pass


class FrequencyNotSupported(ValueError):
    """Error raised when the requested frequency is neither 1Hz nor 20Hz, the
    only frequencies available in the input files."""
    def __init__(self, frequency: int) -> None:
        """"""
        self.frequency = frequency
        msg = 'This reader only supports 1Hz and 20Hz data'
        super(FrequencyNotSupported, self).__init__(msg)


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',
           '    frequency\tEither 1 or 20 (1Hz by default)',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def mask_ssha(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for SSHA.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the SSHA variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 50m).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 50.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_swh(var_dicts: typing.Dict[str, typing.Any],
             var_id: str
             ) -> numpy.typing.NDArray:
    """Build mask for SWH.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the SWH variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 50m).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 50.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_sig0(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for Sigma0.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the Sigma0 variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (values above 100).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (values > 100.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_wind(var_dicts: typing.Dict[str, typing.Any],
              var_id: str
              ) -> numpy.typing.NDArray:
    """Build mask for Wind speed.

    Parameters
    ----------
    var_dict: dict
        Dictionary describing the Wind speed variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. NaN
        or aberrant values (absolute values above 200m.s-1).
    """
    var_dict = var_dicts[var_id]
    values = var_dict['array']
    values._sharedmask = False
    nan_mask = (numpy.isnan(values))
    values[numpy.where(nan_mask)] = var_dict['_FillValue']  # avoid warning
    out_of_range_mask = (numpy.absolute(values) > 200.)
    _mask = (nan_mask | out_of_range_mask)
    result: numpy.typing.NDArray = numpy.ma.getdata(_mask)
    return result


def mask_rain(rain: numpy.typing.NDArray
              ) -> numpy.typing.NDArray:
    """Build mask using Rain flag for Wind, SSH and Sigma0.

    Parameters
    ----------
    rain: numpy.ndarray
        Rain flag variable

    Returns
    -------
    numpy.ndarray
        Array with cells set to True where values should be masked , i.e. Rain
        or high probability of rain from altimeter.
    """
    # build rain flag from bits
    RAIN_BITS = {'no_rain': 0,
                 'rain': 1,
                 'high_rain_probability from_altimeter': 2,
                 'high_probabilty_of_no_rain_from_altimeter': 3,
                 'ambiguous': 4}

    b_rain = numpy.int8(0)
    for frain in ('rain', 'high_rain_probability from_altimeter'):
        b_rain = b_rain + (1 << RAIN_BITS[frain])
    result = (numpy.bitwise_and(rain, b_rain) > 0)
    return result


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    variables = {'wind_speed_alt': {'name': 'wind_speed',
                                    'long_name': 'Wind speed',
                                    'units': 'm.s-1',
                                    'valid_min': 0.0,
                                    'valid_max': 25.4,
                                    'options': {}},
                 'ssha': {'name': 'ssha',
                          'long_name': 'Sea surface height anomaly',
                          'units': 'm',
                          'valid_min': -1.0,
                          'valid_max': 1.0,
                          'options': {}},
                 'swh_ocean': {'name': 'swh',
                               'long_name': 'Significant wave height',
                               'units': 'm',
                               'valid_min': 0.0,
                               'valid_max': 25.0,
                               'options': {}},
                 'sig0_ocean': {'name': 'sigma0',
                                'long_name': 'Sigma0',
                                'units': '',
                                'valid_min': 5.0,
                                'valid_max': 30.0,
                                'options': {}}}

    idf_version = output_opts.get('idf_version', '1.0')

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    freq_hz = int(input_opts.get('frequency', 1))
    if freq_hz not in (1, 20):
        raise FrequencyNotSupported(freq_hz)

    lon_name = 'longitude'
    lat_name = 'latitude'
    time_name = 'time'
    surface_name = 'surface_classification_flag'
    mask_methods = {'ssha': mask_ssha,
                    'swh_ocean': mask_swh,
                    'sig0_ocean': mask_sig0,
                    'wind_speed_alt': mask_wind}

    values: typing.Dict[str, numpy.typing.NDArray] = {}
    get_netcdf_values = idf_converter.lib.extract_variable_values

    f_handler = netCDF4.Dataset(input_path, 'r')
    d_group = f'data_{freq_hz:02d}'
    d_handler = f_handler[d_group]
    if 1 == freq_hz:
        spatial_res_meters = 7000  # very approximative
        time = get_netcdf_values(d_handler, time_name)
        lon = get_netcdf_values(d_handler, lon_name)
        lat = get_netcdf_values(d_handler, lat_name)
        surface = get_netcdf_values(d_handler, surface_name)
        rms = get_netcdf_values(d_handler['ku'], 'swh_ocean_rms')
        rain = get_netcdf_values(d_handler, 'rain_flag')
        rain_mask = mask_rain(rain)
    elif 20 == freq_hz:
        _ = variables.pop('wind_speed_alt')
        spatial_res_meters = 350  # very approximative
        time = get_netcdf_values(d_handler['ku'], time_name)
        lon = get_netcdf_values(d_handler['ku'], lon_name)
        lat = get_netcdf_values(d_handler['ku'], lat_name)
        surface = get_netcdf_values(d_handler['ku'], surface_name)

    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables
    for var_id in variables.keys():
        if 'wind' in var_id:
            idf_converter.lib.extract_variable_attributes(d_handler, var_id,
                                                          granule)
            band = get_netcdf_values(d_handler, var_id)
        else:
            idf_converter.lib.extract_variable_attributes(d_handler['ku'],
                                                          var_id,
                                                          granule)
            band = get_netcdf_values(d_handler['ku'], var_id)

        if (('sigma0' in var_id or 'swh' in var_id) and (freq_hz == 1)):
            _name = f'{var_id}_qual'
            mask = get_netcdf_values(d_handler['ku'], _name)
            band[numpy.where(mask == 1)] = numpy.nan

        values[var_id] = band

    idf_converter.lib.extract_global_attributes(f_handler, input_opts,
                                                granule)
    f_handler.close()

    # Trick to deal with continuity in longitude
    lref = lon[int(0.5 * numpy.shape(lon)[0])]
    lon = numpy.mod(lon - (lref - 180.), 360.) + (lref - 180.)
    lon = numpy.rad2deg(numpy.unwrap(numpy.deg2rad(lon)))

    granule.vars['lat'] = {'array': None,
                           'units': 'degrees north',
                           'datatype': lat.dtype,
                           'options': {}}

    granule.vars['lon'] = {'array': None,
                           'units': 'degrees east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Apply time offset to transform time into an EPOCH timestamp
    ref_time = datetime.datetime(2000, 1, 1)
    epoch = datetime.datetime(1970, 1, 1)
    time_offset = (ref_time - epoch).total_seconds()
    shifted_time = (time + time_offset).astype(numpy.double)

    time_units = 'seconds since 1970-01-01T00:00:00.000Z'
    granule.vars['time'] = {'array': None,
                            'units': time_units,
                            'datatype': numpy.double,
                            'options': {}}

    channels = list(values.keys())
    _granule_name = os.path.basename(os.path.dirname(input_path))
    granule_name = f'{_granule_name}_{freq_hz:02d}hz'
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = spatial_res_meters
    granule.meta['idf_spatial_resolution_units'] = 'm'

    for var_id, band in values.items():
        granule.vars[var_id]['array'] = band
        granule.vars[var_id]['datatype'] = band.dtype

    granule.vars['lat']['array'] = lat
    granule.vars['lon']['array'] = lon
    granule.vars['time']['array'] = shifted_time

    granule.dims['time'] = numpy.size(granule.vars['time']['array'])

    start_dt = datetime.datetime.utcfromtimestamp(shifted_time[0])
    stop_dt = datetime.datetime.utcfromtimestamp(shifted_time[-1])

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt

    transforms = []

    surface_mask = (surface > 0)
    transforms.append(('static_common_mask', {'targets': channels,
                                              'mask': surface_mask}))
    if freq_hz == 1:
        rms_mask = (rms > 1)
        transforms.append(('static_common_mask', {'targets': channels,
                                                  'mask': rms_mask}))

        channels_rain = [item for item in channels if 'swh' not in item]
        transforms.append(('static_common_mask', {'targets': channels_rain,
                                                  'mask': rain_mask}))

    transforms.append(('remove_extra_lon_degrees', {'lon_name': 'lon'}))

    transforms.append(('mask_methods', {'targets': channels,
                                        'methods': mask_methods}))

    output_opts['__export'] = channels

    # data_model, dims, vars, attrs, formatter_jobs
    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains methods for removing signal portions generated by the
traversal of atmosphere."""

import numpy
import typing
import logging
import scipy.interpolate
from idf_converter.lib.types import AtmCorrData, FormatterResult
from idf_converter.lib.types import InputOptions, OutputOptions, Granule

logger = logging.getLogger(__name__)


class NonFiniteValueFound(ValueError):
    """Error raised when an infinite or a NaN value is present in the
    atmospheric correction, most likely due to invalid values in the lookup
    table."""
    def __init__(self) -> None:
        """"""
        _msg = ('Infinite or nan value found in the atmospheric correction, '
                'please check that the LUT has valid values for the angles '
                'contained in the input data file')
        super(NonFiniteValueFound, self).__init__(_msg)


class IncompatibleLUTSize(ValueError):
    """Error raised when the number of entries in the LUT is not a multiple of
    (number of observation zenith angles * number of zenithal deltas)."""
    def __init__(self, lut_size: int, lut_size_factor: float) -> None:
        """"""
        self.lut_size = lut_size
        self.lut_size_factor = lut_size_factor
        _msg = ('The size of the LUT does not match the hardcoded '
                'parameters: the number of lines per sza value should '
                f'be {lut_size_factor}')
        super(IncompatibleLUTSize, self).__init__(_msg)


def s3_cnes_rayleigh(input_opts: InputOptions,
                     output_opts: OutputOptions,
                     granule: Granule,
                     *args: typing.Iterable[typing.Any],
                     targets: typing.Iterable[str],
                     atm_corr_data: AtmCorrData) -> FormatterResult:
    """Apply correction for Rayleigh scattering based on lookup tables
    generated by CNES.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables on which the mask will be applied
    atm_corr_data: dict
        Dictionary containing all the variables and parameters required for
        computing the correction

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the Rayleigh scattering
          has been removed from the targetted variables
    """
    # LUT data
    sza_lut = atm_corr_data['SZA_LUT']
    oza_lut = atm_corr_data['OZA_LUT']
    delta_lut = atm_corr_data['DELTA_LUT']
    rho_atm = atm_corr_data['RHO_LUT']

    # Granule angles
    sza = atm_corr_data['SZA']  # Sun Zenithal Angle
    saa = atm_corr_data['SAA']  # Sun Azimuthal Angle
    oaa = atm_corr_data['OAA']  # Observation Azimuthal Angle
    oza = atm_corr_data['OZA']  # Observation Zenithal Angle

    # Extra-terrestrial solar irradiance
    solar_flux = atm_corr_data['solar_flux']

    bands_count = atm_corr_data['bands_count']
    bands_index = atm_corr_data['bands_index']
    ac_subsampling = atm_corr_data['ac_subsampling']
    al_subsampled = atm_corr_data['al_subsampled']
    nrow = atm_corr_data['nrow']
    ncell = atm_corr_data['ncell']

    oza_resol = 5
    delta_resol = 10
    ntot_delta = 360 / delta_resol + 1
    ntot_oza = 90 / oza_resol + 1

    # Check the size of the LUT
    lut_size = len(rho_atm)
    lut_size_factor = ntot_oza * ntot_delta
    if 0 != (lut_size % lut_size_factor):
        raise IncompatibleLUTSize(lut_size, lut_size_factor)

    # Delta = Sun Azimuthal Angle - Observation Azimuthal Angle
    delta = saa - oaa
    delta = (delta + 360) % 360

    # Interpolate atmospheric correction
    nrow_geom = numpy.shape(sza)[0]
    ncell_geom = numpy.shape(sza)[1]

    rho = numpy.zeros((bands_count, nrow, ncell))
    _points = numpy.array((sza.flatten(), oza.flatten(), delta.flatten())
                          ).transpose()
    sza_lut = numpy.unique(numpy.array(sza_lut, dtype='float32'))
    oza_lut = numpy.unique(numpy.array(oza_lut, dtype='float32'))
    delta_lut = numpy.unique(numpy.array(delta_lut, dtype='float32'))
    lut_points = (sza_lut, oza_lut, delta_lut)
    rho_atm = numpy.array(rho_atm[:][:])
    lut_shape = (sza_lut.size, oza_lut.size, delta_lut.size, 1)

    dst_range_ac = numpy.arange(ncell)
    interpolator = scipy.interpolate.RegularGridInterpolator

    al_rows = nrow_geom
    if al_subsampled is True:
        # Avoid issues in case geometry has more rows than expected (might
        # happen in some edge case)
        al_rows = numpy.minimum(nrow_geom,
                                numpy.floor(nrow / 2.0)).astype('int')
    for var_id in targets:
        iband = bands_index[var_id]

        lut_values = numpy.array((rho_atm[:, iband])).reshape(lut_shape)
        lut_func = interpolator(lut_points, lut_values, method='linear',
                                bounds_error=False, fill_value=None)

        _values = lut_func(_points)
        _values = _values.reshape(nrow_geom, ncell_geom)
        _values = _values * numpy.cos(sza * numpy.pi / 180.)

        src_range_ac = numpy.arange(_values.shape[1]) * ac_subsampling
        # Compensate for along track subsampling (no interpolation along track)
        for irow in range(0, al_rows):
            src_values = _values[irow]

            if al_subsampled is True:
                # Right now the only example we have is Sentinel-3 SLSTR which
                # has a fixed /2 subsampling.
                rho[iband, irow * 2] = numpy.interp(dst_range_ac, src_range_ac,
                                                    src_values)
                irownext = irow * 2 + 1
                if irownext >= nrow:
                    break
                rho[iband, irownext] = numpy.interp(dst_range_ac, src_range_ac,
                                                    src_values)
            else:
                # TODO: check if the "irow < nrow" check is required here, it
                # was in the original OLCI L2 reader
                if irow >= nrow:
                    break

                rho[iband, irow] = numpy.interp(dst_range_ac, src_range_ac,
                                                src_values)

        # TODO check if there are still nan
        rho[numpy.where(numpy.isnan(rho))] = 0

        # 10**(-3) variation for solar flux is neglectable so we consider the
        # mean
        es = numpy.mean(solar_flux[var_id])

        # Convert reflectance into radiance TOA:
        toa_rad = rho[iband, :, :] * es / numpy.pi

        # An erroneous LUT could produce nan values for the atmospheric
        # correction. This is not acceptable.
        if numpy.any(~numpy.isfinite(toa_rad)):
            raise NonFiniteValueFound()

        # Apply correction
        result = granule.vars[var_id]['array']
        corrected = result - toa_rad
        granule.vars[var_id]['array'] = corrected

    return (input_opts, output_opts, granule)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module provides methods to filter data matrices.
"""

import typing
import logging
import scipy.signal
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult, NumValuesDict, Extrema

logger = logging.getLogger(__name__)

MaskGetter = typing.Callable[[Granule, typing.List[str]], NumValuesDict]
ExtremaMethod = typing.Callable[[typing.Dict[str, typing.Any], float, float],
                                Extrema]


def median(input_opts: InputOptions,
           output_opts: OutputOptions,
           granule: Granule,
           *args: typing.Iterable[typing.Any],
           targets: typing.Iterable[str],
           kernels: typing.Dict[str, typing.List[int]]) -> FormatterResult:
    """
    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables that must be converted to logarithmic scale

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been converted to logarithmic scale
    """
    for var_id in targets:
        if var_id not in kernels:
            continue

        kernel = kernels[var_id]
        var_data = granule.vars[var_id]['array']
        var_data = scipy.signal.medfilt(var_data, kernel_size=kernel)
        granule.vars[var_id]['array'] = var_data
        # (3, 3))

    return (input_opts, output_opts, granule)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import numpy
import numpy.typing
import typing
import logging
import netCDF4
import datetime
import idf_converter.lib.constants
import idf_converter.lib.interp
from idf_converter.lib.geo import _fix_swath_longitudes, fix_swath_longitudes
from idf_converter.lib.types import InputOptions, OutputOptions, ProcessedLists
from idf_converter.lib.types import Granule, FormatterResult, GCPs

logger = logging.getLogger(__name__)


def help() -> typing.Tuple[str, str]:
    """"""
    inp = ('    geoloc_spacing\tTODO',
           '    geoloc_row_spacing\tTODO',
           '    geoloc_cell_spacing\tTODO',
           '    geoloc_at_pixel_center\tBoolean set to true if the coordinates'
           ' provided in the input file correspond to the center of the pixels'
           ' and to false if they are referencing the top-left corner of the'
           ' pixels',
           )
    out = ('    gcp_spacing\tTODO',
           '    gcp_row_spacing\tTODO',
           '    gcp_cell_spacing\tTODO',
           '    gcp_distribution\tTODO',
           )
    return ('\n'.join(inp), '\n'.join(out))


def preformat(input_opts: InputOptions, output_opts: OutputOptions,
              granule: Granule
              ) -> typing.Iterator[FormatterResult]:
    """"""
    yield input_opts, output_opts, granule


def _get_distribution_arity(row_count: int, cell_count: int, row_spacing: int,
                            cell_spacing: int) -> typing.Tuple[int, int]:
    """"""
    # Number of subdivisions along each axis
    row_spaces = row_count - 1
    cell_spaces = cell_count - 1

    # Compute the number of GCPs to place along each axis of the geolocation
    # matrix so that the distance in row/cell between two adjacent GCPs is
    # inferior or equal to the requested spacing
    _output_row_spaces = row_spaces / row_spacing
    output_row_spaces = numpy.ceil(_output_row_spaces).astype('int')
    _output_cell_spaces = cell_spaces / cell_spacing
    output_cell_spaces = numpy.ceil(_output_cell_spaces).astype('int')

    # Convert from number of subdivisions to number of row/cell
    output_nrow = output_row_spaces + 1
    output_ncell = output_cell_spaces + 1

    return (output_nrow, output_ncell)


def create_uniform_gcp_distribution(granule: Granule,
                                    input_opts: InputOptions,
                                    output_opts: OutputOptions,
                                    geo_shape: numpy.typing.NDArray
                                    ) -> typing.Dict[str,
                                                     numpy.typing.NDArray]:
    """"""
    full_geoloc_nrow = geo_shape[0]
    full_geoloc_ncell = geo_shape[1]
    data_nrow = granule.dims['row']
    data_ncell = granule.dims['cell']

    # Number of digital points per GCP in the input
    geoloc_spacing = int(input_opts.get('geoloc_spacing', 32))
    geoloc_row_spacing = int(input_opts.get('geoloc_row_spacing',
                                            geoloc_spacing))
    geoloc_cell_spacing = int(input_opts.get('geoloc_cell_spacing',
                                             geoloc_spacing))

    geoloc_nrow, geoloc_ncell = _get_distribution_arity(full_geoloc_nrow,
                                                        full_geoloc_ncell,
                                                        geoloc_row_spacing,
                                                        geoloc_cell_spacing)

    # Compute uniform distribution for geoloc data
    # The max index value (array length - 1) MUST be included in the result
    _geoloc_row = numpy.linspace(0, full_geoloc_nrow - 1, num=geoloc_nrow)
    geoloc_row = numpy.round(_geoloc_row).astype('int')
    _geoloc_cell = numpy.linspace(0, full_geoloc_ncell - 1, num=geoloc_ncell)
    geoloc_cell = numpy.round(_geoloc_cell).astype('int')

    # Number of digital points per GCP in the output
    gcp_spacing = int(output_opts.get('gcp_spacing', 32))
    gcp_row_spacing = int(output_opts.get('gcp_row_spacing',
                                          gcp_spacing))
    gcp_cell_spacing = int(output_opts.get('gcp_cell_spacing',
                                           gcp_spacing))

    # Warn user if requested GCP resolution exceeds geolocation resolution for
    # the row axis
    _native_row_spacing = data_nrow / full_geoloc_nrow
    native_row_spacing = numpy.ceil(_native_row_spacing).astype('int')
    if native_row_spacing > gcp_row_spacing:
        logger.warning('Requested GCP spacing on the row axis is lower than '
                       'geolocation spacing in input data')
        logger.debug(f'Native geoloc row spacing: {native_row_spacing}')
        logger.debug(f'Requested row spacing: {gcp_row_spacing}')

    # Warn user if requested GCP resolution exceeds geolocation resolution for
    # the cell axis
    _native_cell_spacing = data_ncell / full_geoloc_ncell
    native_cell_spacing = numpy.ceil(_native_cell_spacing).astype('int')
    if native_cell_spacing > gcp_cell_spacing:
        logger.warning('Requested GCP spacing on the cell axis is lower than '
                       'geolocation spacing in input data')
        logger.debug(f'Native geoloc cell spacing: {native_cell_spacing}')
        logger.debug(f'Requested cell spacing: {gcp_cell_spacing}')

    # GCPs will be placed automatically on the bottom and right edges of the
    # data matrix at a later stage, so consider that the data matrix is shorter
    # by one requested spacing along each axis
    tmp_nrow = data_nrow - gcp_row_spacing
    tmp_ncell = data_ncell - gcp_cell_spacing

    # Number of output GCPs to place along each direction
    gcp_nrow, gcp_ncell = _get_distribution_arity(tmp_nrow, tmp_ncell,
                                                  gcp_row_spacing,
                                                  gcp_cell_spacing)

    # Compute uniform distribution for output GCPs
    gcp_row = numpy.arange(gcp_nrow) * gcp_row_spacing
    gcp_cell = numpy.arange(gcp_ncell) * gcp_cell_spacing

    # Compute indices in the data matrix that match the extracted geoloc inputs
    row = numpy.linspace(0, data_nrow - 1, num=full_geoloc_nrow)
    cell = numpy.linspace(0, data_ncell - 1, num=full_geoloc_ncell)
    data_row = row[geoloc_row]
    data_cell = cell[geoloc_cell]

    result = {'geoloc_row': geoloc_row, 'geoloc_cell': geoloc_cell,
              'input_row': data_row, 'input_cell': data_cell,
              'output_row': gcp_row, 'output_cell': gcp_cell}

    return result


def compute_gcps(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 granule: Granule) -> GCPs:
    """"""
    nrow = granule.dims['row']
    ncell = granule.dims['cell']

    # Compute GCPs
    lon = granule.vars['lon']['array']
    lat = granule.vars['lat']['array']

    gcp_distribution = output_opts.get('gcp_distribution', None)
    if gcp_distribution is None:
        gcp_distribution = create_uniform_gcp_distribution(granule,
                                                           input_opts,
                                                           output_opts,
                                                           numpy.shape(lon))

    _geoloc_row = gcp_distribution['geoloc_row']
    _geoloc_cell = gcp_distribution['geoloc_cell']
    _input_row = gcp_distribution['input_row']
    _input_cell = gcp_distribution['input_cell']
    _output_row = gcp_distribution['output_row']
    _output_cell = gcp_distribution['output_cell']

    geoloc_nrow = numpy.shape(_geoloc_row)[0]
    geoloc_ncell = numpy.shape(_geoloc_cell)[0]

    # Indices of the data matrix must be reshaped so that numpy can use them
    # to extract the associated values
    let_numpy_guess = -1
    geoloc_row = _geoloc_row.reshape((let_numpy_guess, 1))
    geoloc_cell = _geoloc_cell.reshape((1, let_numpy_guess))

    # Read geographical coordinates of the reference points that will be used
    # compute the GCPs
    src_lon = lon[geoloc_row, geoloc_cell]
    src_lat = lat[geoloc_row, geoloc_cell]

    # GCP pixels are located at the edge of data pixels, meaning that the
    # center of the first data pixel is located at (0.5, 0.5) in the GCPs
    # matrix.
    # If lon/lat correspond to the center of data pixels, then the indices
    # stored in the GCPs must be shifted by one half-pixel along each axis
    _geoloc_at_pixel_center = input_opts.get('geoloc_at_pixel_center', 'true')
    geoloc_at_pixel_center = _geoloc_at_pixel_center.lower() in ('true', 'yes')
    if geoloc_at_pixel_center:
        _input_row = _input_row + 0.5
        _input_cell = _input_cell + 0.5

    # Build a matrix for the row and cell coordinates of the reference points
    # in the data matrix by repeating indices vector along the orthogonal axis
    input_row = _input_row.reshape((let_numpy_guess, 1))
    input_cell = _input_cell.reshape((1, let_numpy_guess))
    input_row = numpy.tile(input_row, (1, geoloc_ncell))
    input_cell = numpy.tile(input_cell, (geoloc_nrow, 1))

    # Indices of resulting GCPs must be integer values
    output_row = numpy.floor(_output_row).astype('int')
    output_cell = numpy.floor(_output_cell).astype('int')

    # Add an extra line and column to geolocate the bottom and right edges of
    # the data matrix
    output_row = numpy.concatenate((output_row, [nrow]))
    output_cell = numpy.concatenate((output_cell, [ncell]))
    ngcps = [output_row.size, output_cell.size]

    # Since the (i, j) of the original GCPs might have been float values, it
    # is necessary to compute the geographical coordinates for the integer
    # version of (i, j)
    _use_gdal = output_opts.get('use_gdal', 'no')
    use_gdal = _use_gdal.lower() in ('true', 'yes')
    if use_gdal is True:
        import warnings
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            try:
                from osgeo import gdal  # find a way to remove this
            except ImportError:
                import gdal  # find a way to remove this
        drv = gdal.GetDriverByName('MEM')
        dset = drv.Create('tmp', ncell, nrow)
        gcps = []
        for _lon, _lat, _cell, _row in zip(src_lon.flat, src_lat.flat,
                                           input_cell.flat, input_row.flat):
            out_lon = float(_lon)
            out_lat = float(_lat)
            out_hei = float(0)
            out_cell = float(_cell)
            out_row = float(_row)
            gcps.append(gdal.GCP(out_lon, out_lat, out_hei, out_cell, out_row))
        dset.SetGCPs(gcps, idf_converter.lib.constants.EPSG4326_WKT)
        options = ['MAX_GCP_ORDER=-1']
        transformer = gdal.Transformer(dset, None, options)
        _output_row = numpy.tile(output_row[:, numpy.newaxis], (1, ngcps[1]))
        _output_cell = numpy.tile(output_cell[numpy.newaxis, :], (ngcps[0], 1))

        # Extra/Interpolate geo coordinates for the requested indices
        inputs = numpy.vstack((_output_cell.flatten(),
                               _output_row.flatten())).transpose()
        outputs = numpy.array(transformer.TransformPoints(0, inputs)[0])[:,
                                                                         0:2]
        output_lat = outputs[:, 1].reshape(ngcps)
        output_lon = outputs[:, 0].reshape(ngcps)
        dset = None
    else:
        interp_gcps = idf_converter.lib.interp.geoloc_from_gcps
        _output_row = numpy.tile(output_row[:, numpy.newaxis], (1, ngcps[1]))
        _output_cell = numpy.tile(output_cell[numpy.newaxis, :], (ngcps[0], 1))
        output_lon, output_lat = interp_gcps(src_lon, src_lat,
                                             input_row, input_cell,
                                             _output_row, _output_cell)
        output_lon = _fix_swath_longitudes(output_lon, 9)

    return output_lon, output_lat, output_row, output_cell


def format_granule(input_opts: InputOptions,
                   output_opts: OutputOptions,
                   granule: Granule,
                   idf_file: netCDF4.Dataset) -> ProcessedLists:
    """Create dimensions, variables and attributes required for swath granules.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    idf_file: netCDF4.Dataset
        Handle of the IDF file to edit

    Returns
    -------
    tuple
        Information about granule information that the method has processed.
        The tuple contains three lists::
        - a list of names for the dimensions created by the method
        - a list of identifiers for the variables created by the method
        - a list of names for the attributes created by the method
    """

    input_opts, output_opts, granule = fix_swath_longitudes(input_opts,
                                                            output_opts,
                                                            granule,
                                                            targets=('lon',),
                                                            decimals=9)

    gcp_lon, gcp_lat, gcp_lin, gcp_pix = compute_gcps(input_opts, output_opts,
                                                      granule)
    gcp_nrow, gcp_ncell = gcp_lon.shape

    nrow = granule.dims['row']
    ncell = granule.dims['cell']

    # Central datetime
    _start_dt = granule.meta['time_coverage_start']
    _stop_dt = granule.meta['time_coverage_end']
    mid_time = _start_dt + (_stop_dt - _start_dt) / 2

    # Precision set to seconds
    microseconds = mid_time.microsecond
    if microseconds > 500000:
        mid_time = mid_time + datetime.timedelta(seconds=1)
    mid_time = mid_time.replace(microsecond=0)

    # Dimensions
    idf_file.createDimension('time', None)  # unlimited
    idf_file.createDimension('row', nrow)
    idf_file.createDimension('cell', ncell)
    idf_file.createDimension('row_gcp', gcp_nrow)
    idf_file.createDimension('cell_gcp', gcp_ncell)
    idf_dims = ('time', 'row', 'cell', 'row_gcp', 'cell_gcp',)
    extra_dims = [x for x in granule.dims.keys() if x not in idf_dims]
    for extra_dim in extra_dims:
        idf_file.createDimension(extra_dim, granule.dims[extra_dim])

    # Variables
    _time = idf_file.createVariable('time', 'f8', ('time', ))
    _time.long_name = 'time'
    _time.standard_name = 'time'
    _time.units = 'seconds since 1970-01-01T00:00:00.000000Z'
    _time.calendar = 'standard'
    _time[:] = (mid_time - datetime.datetime(1970, 1, 1)).total_seconds()

    _latgcp = idf_file.createVariable('lat_gcp', 'f4',
                                      ('row_gcp', 'cell_gcp'))
    _latgcp.long_name = 'ground control points latitude'
    _latgcp.standard_name = 'latitude'
    _latgcp.units = 'degrees_north'
    _latgcp[:] = gcp_lat.astype('float32')

    _longcp = idf_file.createVariable('lon_gcp', 'f4', ('row_gcp', 'cell_gcp'))
    _longcp.long_name = 'ground control points longitude'
    _longcp.standard_name = 'longitude'
    _longcp.units = 'degrees_east'
    _longcp[:] = gcp_lon.astype('float32')

    _indexrowgcp = idf_file.createVariable('index_row_gcp', 'i4', ('row_gcp',))
    _indexrowgcp.long_name = ('index of ground control points in row '
                              'dimension')
    _indexrowgcp.comment = ('index goes from 0 (start of first pixel) to '
                            'dimension value (end of last pixel)')
    _indexrowgcp[:] = gcp_lin.astype('int32')

    _indexcellgcp = idf_file.createVariable('index_cell_gcp', 'i4',
                                            ('cell_gcp',))
    _indexcellgcp.long_name = ('index of ground control points in cell '
                               'dimension')
    _indexcellgcp.comment = ('index goes from 0 (start of first pixel) to '
                             'dimension value (end of last pixel)')
    _indexcellgcp[:] = gcp_pix.astype('int32')

    idf_file.cdm_data_type = b'Swath'

    processed_dims = ['time', 'row', 'cell', 'row_gcp', 'cell_gcp']
    processed_vars = ['time', 'lat_gcp', 'lon_gcp', 'index_row_gcp',
                      'index_cell_gcp', 'lat', 'lon']
    processed_attrs = ['cdm_data_type']

    return (processed_dims, processed_vars, processed_attrs)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains formatter methods for operations related to data
masking.
"""

import numpy
import numpy.typing
import typing
import logging
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult, NumValuesDict

logger = logging.getLogger(__name__)

MaskMethod = typing.Callable[[typing.Dict[str, typing.Any], str],
                             numpy.typing.NDArray]


def static_common_mask(input_opts: InputOptions,
                       output_opts: OutputOptions,
                       granule: Granule,
                       *args: typing.Iterable[typing.Any],
                       targets: typing.Iterable[str],
                       mask: numpy.typing.NDArray) -> FormatterResult:
    """
    Apply a common mask on one or several variables.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables on which the mask will be applied
    mask: numpy.ndarray
        Array of boolean values used to mask variables. Any cell set to True in
        this array will be masked

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been masked
    """
    masked_ind = numpy.where(mask)
    for var_name in targets:
        granule.vars[var_name]['array'][masked_ind] = numpy.nan

    return input_opts, output_opts, granule


def static_bands_mask(input_opts: InputOptions,
                      output_opts: OutputOptions,
                      granule: Granule,
                      *args: typing.Iterable[typing.Any],
                      targets: typing.Iterable[str],
                      masks: NumValuesDict) -> FormatterResult:
    """
    Apply variable-specific masks.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables on which the mask will be applied
    masks: numpy.ndarray
        Array of boolean values used to mask variables. Any cell set to True in
        this array will be masked

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been masked
    """
    for var_id in targets:
        if var_id not in masks.keys():
            continue
        masked_ind = numpy.where(masks[var_id])
        granule.vars[var_id]['array'][masked_ind] = numpy.nan

    return input_opts, output_opts, granule


def mask_methods(input_opts: InputOptions,
                 output_opts: OutputOptions,
                 granule: Granule,
                 *args: typing.Iterable[typing.Any],
                 targets: typing.Iterable[str],
                 methods: typing.Dict[str, MaskMethod]) -> FormatterResult:
    """
    Build and apply a custom mask on each targetted variable. The methods for
    building the mask must accept an entry from the Granule.vars dictionary as
    input.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    targets: tuple
        Identifers for variables on which the mask will be applied
    methods: dict
        Dictionary with variable identifiers as keys and methods for building
        custom masks as values

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the targetted variables
          have been masked
    """
    for var_id in targets:
        var_mask = methods[var_id](granule.vars, var_id)
        masked_ind = numpy.where(var_mask)
        granule.vars[var_id]['array'][masked_ind] = numpy.nan

    return input_opts, output_opts, granule


def share_masks(input_opts: InputOptions,
                output_opts: OutputOptions,
                granule: Granule,
                *args: typing.Iterable[typing.Any],
                targets: typing.Iterable[str]) -> FormatterResult:
    """"""
    all_masks = [numpy.ma.getmaskarray(granule.vars[var_id]['array'])
                 for var_id in targets]
    common_mask = numpy.any(all_masks, axis=0)
    for var_id in targets:
        granule.vars[var_id]['array'].mask = common_mask

    return input_opts, output_opts, granule



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import typing
import netCDF4
import logging
import idf_converter.lib
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger(__name__)

DATA_MODEL = 'SWATH'


class InputPathMissing(IOError):
    """"""
    pass


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    variables = ('wind_speed', 'wind_dir', 'model_speed', 'model_dir')
    values = {}
    f_handler = netCDF4.Dataset(input_path, 'r')
    for var_id in variables:
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        _band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        values[var_id] = numpy.ma.array(_band)
    lon = idf_converter.lib.extract_variable_values(f_handler, 'lon')
    lat = idf_converter.lib.extract_variable_values(f_handler, 'lat')
    _time = idf_converter.lib.extract_variable_values(f_handler, 'time')
    time_units = f_handler.variables['time'].units
    platform, sensor = f_handler.source.split(' ')
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    nrow = f_handler.dimensions['NUMROWS'].size
    ncell = f_handler.dimensions['NUMCELLS'].size
    f_handler.close()

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)

    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 12500
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['platform'] = platform
    granule.meta['sensor'] = sensor

    half_cells = numpy.floor(ncell / 2).astype('int')
    sides = {'left': slice(0, half_cells, None),
             'right': slice(half_cells, None, None)}
    for side, cell_slice in sides.items():
        granule.meta['idf_granule_id'] = f'{granule_name}_{side}'

        granule.dims['row'] = int(nrow)
        granule.dims['cell'] = int(half_cells)
        slices = (slice(None, None, None), cell_slice)

        # Format spatial information
        granule.vars['lat'] = {'array': lat[slices],
                               'units': 'degrees_north',
                               'datatype': lat.dtype,
                               'options': {}}
        granule.vars['lon'] = {'array': lon[slices],
                               'units': 'degrees_east',
                               'datatype': lon.dtype,
                               'options': {}}

        # Format time information
        time_min = _time[slices].min()
        time_max = _time[slices].max()
        start_dt = netCDF4.num2date(time_min, units=time_units)
        stop_dt = netCDF4.num2date(time_max, units=time_units)
        granule.meta['time_coverage_start'] = start_dt
        granule.meta['time_coverage_end'] = stop_dt

        for var_id in values:
            granule.vars[var_id] = {'array': values[var_id][slices]}

        # Add transforms
        transforms = []

        dir_cfg = {'wind': {'direction': 'wind_dir',
                            'module': 'wind_speed',
                            'radians': False,
                            'meteo': False,
                            'angle_to_east': numpy.pi / 2.0,
                            'clockwise': True},
                   'ecmwf': {'direction': 'model_dir',
                             'module': 'model_speed',
                             'radians': False,
                             'meteo': False,
                             'angle_to_east': numpy.pi / 2.0,
                             'clockwise': True}}
        transforms.append(('dir2vectors', {'targets': ('wind', 'ecmwf'),
                                           'configs': dir_cfg}))
        channels = ['eastward_wind', 'northward_wind', 'eastward_ecmwf',
                    'northward_ecmwf']

        # Remove direction and modulus variables
        to_remove = ('wind_dir', 'wind_speed', 'model_dir', 'model_speed')
        transforms.append(('remove_vars', {'targets': to_remove}))

        output_opts['__export'] = channels

        yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-27
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import datetime
import idf_converter.lib
import defusedxml.ElementTree as ET
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger(__name__)

MasksDict = typing.Dict[str, numpy.typing.NDArray]
DATA_MODEL = 'SWATH'
MAX_DIM_SIZE = 2**13

# https://earth.esa.int/web/sentinel/technical-guides/sentinel-3-slstr/level-1/vis-and-swir-radiances
# https://sentinel.esa.int/web/sentinel/user-guides/sentinel-3-slstr/data-formats/level-1
# https://earth.esa.int/web/sentinel/user-guides/sentinel-3-slstr/product-types/level-1b
# https://sentinel.esa.int/documents/247904/685236/Sentinel-3_User_Handbook#%5B%7B%22num%22%3A235%2C%22gen%22%3A0%7D%2C%7B%22name%22%3A%22XYZ%22%7D%2C54%2C628%2C0%5D
# S1 555   nm            B
# S2 659   nm            G
# S3 865   nm            R
# S4 1.375 microm
# S5 1.610 microm
# S6 2.25  microm
# S7 3.74  microm        IR
# S8 10.85 microm        IR
# S9 12    microm        IR


class InputPathMissing(IOError):
    """"""
    pass


class TimelinessNotSupported(Exception):
    """Error raised when the input file is a Non Time Critical product: this
    reader only supports Near Real Time products."""
    def __init__(self, timeliness: str) -> None:
        """"""
        self.timeliness = timeliness


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    inp = ('    path\tPath of the input file',)
    out = ('',)
    return ('\n'.join(inp), '\n'.join(out))


def get_timeliness(input_dir: str) -> typing.Optional[str]:
    """ Read metadata in xml file. Return start and track Offset, number of
    lines and elements spatial resolution for Tie Points, 1 km nadir and 1 km
    oblique grid."""
    manifest_path = os.path.join(input_dir, 'xfdumanifest.xml')
    if not os.path.exists(manifest_path):
        return None
    tree = ET.parse(manifest_path)
    root = tree.getroot()
    s3path = '{http://www.esa.int/safe/sentinel/sentinel-3/1.0}'
    for a_child in root.findall("./*/metadataObject/metadataWrap/xmlData/*"):
        for child in a_child:
            if child.tag == f'{s3path}timeliness':
                return str(child.text)
    return None


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """

    variables = {'sea_surface_temperature': {'name': 'sea_surface_temperature',
                                             'valid_min': 270.150,
                                             'valid_max': 310.150,
                                             'options': {}},
                 'wind_speed': {'name': 'wind_speed',
                                'long_name': 'wind speed',
                                'valid_min': 0.000,
                                'valid_max': 25.400,
                                'options': {}}}

    channels = list(variables.keys())

    idf_version = output_opts.get('idf_version', '1.0')
    granule = idf_converter.lib.create_granule(idf_version, DATA_MODEL)
    granule.vars = variables

    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise InputPathMissing()
    input_path = os.path.normpath(_input_path)

    input_dir = os.path.dirname(input_path)
    timeliness = get_timeliness(input_dir)
    logger.debug(f'timeliness: {timeliness}')
    if timeliness is None:
        logger.warning('Could not read data timeliness. Please note that only '
                       'NRT data are supported')
    elif 'NT' == timeliness:
        logger.error('Input data are not supported due to the size of the '
                     'along-track dimension which exceeds SEAScope '
                     'capabilities')
        raise TimelinessNotSupported(timeliness)

    f_handler = netCDF4.Dataset(input_path, 'r')
    for var_id in channels:
        idf_converter.lib.extract_variable_attributes(f_handler, var_id,
                                                      granule)
        _band = idf_converter.lib.extract_variable_values(f_handler, var_id)
        band = _band[0, :, :]
        granule.vars[var_id]['array'] = numpy.ma.array(band)
    lon = idf_converter.lib.extract_variable_values(f_handler, 'lon')
    lat = idf_converter.lib.extract_variable_values(f_handler, 'lat')
    _sst_dtime = idf_converter.lib.extract_variable_values(f_handler,
                                                           'sst_dtime')
    sst_dtime = _sst_dtime[0, :, :]
    _time = idf_converter.lib.extract_variable_values(f_handler, 'time')
    time0 = _time[0]
    time0_units = f_handler.variables['time'].units
    _ql = idf_converter.lib.extract_variable_values(f_handler,
                                                    'quality_level')
    ql = _ql[0, :, :]
    idf_converter.lib.extract_global_attributes(f_handler, input_opts, granule)
    f_handler.close()

    dtime0 = netCDF4.num2date(time0, units=time0_units)

    nrow, ncell = numpy.shape(granule.vars['sea_surface_temperature']['array'])
    granule.dims['row'] = nrow
    granule.dims['cell'] = ncell

    # Handle longitude continuity
    dlon = lon[1:, :] - lon[:-1, :]
    if 180.0 <= numpy.max(numpy.abs(dlon)):
        lon[numpy.where(lon < 0.0)] = lon[numpy.where(lon < 0.0)] + 360.0

    granule.vars['lat'] = {'array': lat,
                           'units': 'degrees_north',
                           'datatype': lat.dtype,
                           'options': {}}
    granule.vars['lon'] = {'array': lon,
                           'units': 'degrees_east',
                           'datatype': lon.dtype,
                           'options': {}}

    # Format time information
    start_dt = dtime0 + datetime.timedelta(seconds=numpy.min(sst_dtime))
    stop_dt = dtime0 + datetime.timedelta(seconds=numpy.max(sst_dtime))

    granule_file = os.path.basename(input_path)
    granule_name, _ = os.path.splitext(granule_file)

    granule.meta['idf_granule_id'] = granule_name
    granule.meta['idf_subsampling_factor'] = 0
    granule.meta['idf_spatial_resolution'] = 1000  # 1km
    granule.meta['idf_spatial_resolution_units'] = 'm'
    granule.meta['time_coverage_start'] = start_dt
    granule.meta['time_coverage_end'] = stop_dt
    granule.meta['institution'] = 'Eumetsat'
    granule.meta['platform'] = 'Sentinel-3'
    granule.meta['platform_type'] = 'leo satellite'
    granule.meta['sensor'] = 'SLSTR'
    granule.meta['sensor_type'] = 'dual scan temperature radiometer'

    # Add transforms
    transforms = []

    # Mask where quality is subpar
    ql_mask = numpy.full(numpy.shape(ql), True, dtype='bool')
    ql_mask[numpy.where(ql > 2)] = False
    transforms.append(('static_common_mask', {'targets': channels.copy(),
                                              'mask': ql_mask}))

    output_opts['__export'] = channels

    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2020-01-16
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""
"""

import os
import numpy
import numpy.typing
import typing
import netCDF4
import logging
import idf_converter.lib
import idf_converter.readers.netcdf_grid_yx
from idf_converter.lib.types import InputOptions, OutputOptions, ReaderResult

logger = logging.getLogger()

DATA_MODEL = idf_converter.readers.netcdf_grid_yx.DATA_MODEL
ValuesAndMasks = typing.Tuple[typing.Dict[str, numpy.typing.NDArray],
                              typing.Dict[str, numpy.typing.NDArray]]


class PlatformMissing(Exception):
    """Error raised when neither the "platform" nor the "platform_id" global
    attributes are available in the input file, preventing the identification
    of the platform (necessary to extract the correct variables and apply the
    adquate masking policy)."""
    pass


class PlatformNotSupported(ValueError):
    """Error raised when the platform referenced in the input file is neither
    QuikSCAT nor one of the MetOp satellites, so the variables to extract are
    unknown, as is the masking policy to apply."""
    def __init__(self, platform: str) -> None:
        """"""
        self.platform = platform


def help() -> typing.Tuple[str, str]:
    """Describe options supported by this reader.

    Returns
    -------
    str
        Description of the supported options
    """
    in_msg, out_msg = idf_converter.readers.netcdf_grid_yx.help()
    return (in_msg, out_msg)


def get_quikscat_mask(ubytes: numpy.typing.NDArray
                      ) -> numpy.typing.NDArray:
    """"""
    mask: numpy.typing.NDArray = ((ubytes == 0) | (ubytes >= 254))

    return mask


def get_metop_mask(ubytes: numpy.typing.NDArray
                   ) -> numpy.typing.NDArray:
    """"""
    mask: numpy.typing.NDArray = ((ubytes == 0) | (ubytes == 255))

    return mask


def _identify_hemisphere(f_handler: netCDF4.Dataset) -> typing.Tuple[str, str]:
    """"""
    # For some reason the latitude variable is entirely msked by default...
    f_handler.variables['latitude'].set_auto_mask(False)
    mean_lat = numpy.mean(f_handler.variables['latitude'][:])

    # The input files contain either the Arctic or the Antarctic region
    if 0.0 < mean_lat:
        projection = 'epsg:3411'
        pole = 'arctic'
    else:
        projection = 'epsg:3412'
        pole = 'antarctic'

    return pole, projection


def _identify_platforms(f_handler: netCDF4.Dataset
                        ) -> typing.Tuple[str, str, typing.List[str]]:
    """"""
    global_attrs = f_handler.ncattrs()
    _platform = None
    if 'platform' in global_attrs:
        _platform = f_handler.platform
    elif 'platform_id' in global_attrs:
        _platform = f_handler.platform_id
    else:
        raise PlatformMissing()

    platform = ''
    sensor = ''
    variables = []
    if 'quikscat' in _platform.lower():
        platform = 'QuikSCAT'
        sensor = 'SeaWinds'
        variables = ['sigma0_mask_inner', 'sigma0_mask_outer']
    elif 'metop' in _platform.lower():
        platforms = []
        if 'b' in _platform.lower():
            platforms.append('MetOp-B')
        else:
            # sometimes MetOp-A is clearly stated, sometimes it is just "MetOp"
            platforms.append('MetOp-A')

        platform = ','.join(platforms)
        sensor = ','.join(['ASCAT'] * len(platforms))
        variables = ['sigma_40_mask', ]
    else:
        raise PlatformNotSupported(_platform)

    return platform, sensor, variables


def _get_values_and_masks(f_handler: netCDF4.Dataset,
                          variables: typing.List[str],
                          platform: str) -> ValuesAndMasks:
    """"""
    # Sigma values have been computed as unsigned bytes with a scale factor and
    # an offset, but have been stored as *signed* bytes without modifying the
    # offset: the upper half of the values range has therefore been shifted to
    # negative values and netCDF4 has no way to guess how these values should
    # be read.
    # The workaround implemented here consists in reading values as unsigned
    # bytes (regardless of what the input file says) before calling the generic
    # grid reader, unpack them correctly and then overwrite the values
    # extracted automatically by the reader.
    values = {}
    masks = {}
    for var_id in variables:
        f_handler.variables[var_id].set_auto_scale(False)
        ubyte_values = f_handler.variables[var_id][:].astype('ubyte')
        f_handler.variables[var_id].set_auto_scale(True)
        if 'QuikSCAT' == platform:
            masks[var_id] = get_quikscat_mask(ubyte_values)
        else:
            masks[var_id] = get_metop_mask(ubyte_values)
        scale = f_handler.variables[var_id].scale_factor
        offset = f_handler.variables[var_id].add_offset
        values[var_id] = numpy.ma.array(offset + scale * ubyte_values)

    return values, masks


def read_data(input_opts: InputOptions,
              output_opts: OutputOptions
              ) -> typing.Iterator[ReaderResult]:
    """Read input file, extract data and metadata, store them in a Granule
    object then prepare formatting instructions to finalize the conversion to
    IDF format.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule, list)
        A tuple which contains four elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` where the extracted information
          has been stored

        - a :obj:list of :obj:dict describing the formatting operations that
          the converter must perform before serializing the result in IDF
          format
    """
    _input_path = input_opts.get('path', None)
    if _input_path is None:
        raise idf_converter.readers.netcdf_grid_yx.InputPathMissing()
    input_path = os.path.normpath(_input_path)

    f_handler = netCDF4.Dataset(input_path, 'r')

    # Detect which hemisphere is contained in the input file
    pole, projection = _identify_hemisphere(f_handler)

    try:
        # Identify the platform(s) and instrument(s) involved in the creation
        # of the input file
        platform, sensor, variables = _identify_platforms(f_handler)

        # Extract sigma values and masks in advance to bypass formatting quirks
        # in the input files
        values, masks = _get_values_and_masks(f_handler, variables, platform)
    finally:
        f_handler.close()

    input_opts['time_variable'] = 'time'
    input_opts['lat_variable'] = 'latitude'
    input_opts['lon_variable'] = 'longitude'
    input_opts['lat_dimension'] = 'nj'
    input_opts['lon_dimension'] = 'ni'
    input_opts['variables'] = ','.join(variables)
    input_opts['time_coverage_relative_start'] = '-43200'
    input_opts['time_coverage_relative_end'] = '43200'
    input_opts['spatial_resolution'] = 12500
    input_opts['projection'] = projection

    output_opts['use_mean_dydx'] = 'true'

    grid_yx = idf_converter.readers.netcdf_grid_yx
    result = grid_yx.read_data(input_opts, output_opts)
    input_opts, output_opts, granule, transforms = next(result)

    granule.meta['platform'] = platform
    granule.meta['sensor'] = sensor

    # For L4 (merged MetOp-A and MetOp-B), the files for Arctic and Antarctic
    # have the exact same name, so they would get the same granule identifier
    # despite being diffrent IDF files. Prefix with the name of the pole to
    # avoid this issue
    granule.meta['idf_granule_id'] = f'{pole}_{granule.meta["idf_granule_id"]}'

    # Overwrite arrays of sigma values since the netcdf_grid_yx reader could
    # not extract the correct values
    for var_id, var_array in values.items():
        granule.vars[var_id]['array'] = var_array

    # Add mask where there is no ice
    for target, mask in masks.items():
        transforms.insert(0, ('static_common_mask', {'targets': (target,),
                                                     'mask': mask}))

    yield (input_opts, output_opts, granule, transforms)



# vim: ts=4:sts=4:sw=4
#
# @date 2019-09-25
#
# This file is part of IDF converter, a set of tools to convert satellite,
# in-situ and numerical model data into Intermediary Data Format, making them
# compatible with the SEAScope application.
#
# Copyright (C) 2014-2022 OceanDataLab
#
# IDF converter is free software: you can redistribute it and/or modify it
# under the terms of the GNU Lesser General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# IDF converter is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with IDF converter. If not, see <https://www.gnu.org/licenses/>.

"""This module contains formatter methods for operations related to
geographical coordinates.
"""

import numpy
import numpy.typing
import typing
import logging
from idf_converter.lib import mod360
from idf_converter.lib.types import InputOptions, OutputOptions, Granule
from idf_converter.lib.types import FormatterResult

logger = logging.getLogger(__name__)


def remove_extra_lon_degrees(input_opts: InputOptions,
                             output_opts: OutputOptions,
                             granule: Granule,
                             *args: typing.Iterable[typing.Any],
                             lon_name: str) -> FormatterResult:
    """
    Shift longitude values to remove extra 360 degrees.

    Parameters
    ----------
    input_opts: dict
        Options and information related to input data
    output_opts: dict
        Options and information related to formatting and serialization to IDF
        format
    granule: idf_converter.lib.Granule
        Object representing the granule data and metadata
    lon_name: str
        Identifier for the variable containing longitudes. It has to be passed
        as a keyword argument.

    Returns
    -------
    tuple(dict, dict, idf_converter.lib.Granule)
        A tuple which contains three elements:

        - the input_options :obj:dict passed to this method

        - the output_options :obj:dict passed to this method

        - the :obj:`idf_converter.lib.Granule` with the shifted longitudes
    """
    lon = granule.vars[lon_name]['array']
    lon = lon - numpy.floor((numpy.min(lon) + 180.) / 360.) * 360.
    granule.vars[lon_name]['array'] = lon
    return (input_opts, output_opts, granule)


def _fix_swath_longitudes(_lon: numpy.typing.NDArray,
                          decimals: int
                          ) -> numpy.typing.NDArray:
    """"""
    # Start by making sure all values are in [-180.0, 180.0[
    lon, trunc_to_0 = mod360(_lon + 180.0, decimals)
    lon = lon - 180.0

    # Find acrosstrack main direction
    ac0_dlon = lon[0, 1:] - lon[0, :-1]
    ac0_dlon, precision_loss = mod360(ac0_dlon, decimals)
    trunc_to_0 = trunc_to_0 or precision_loss
    ac0_dlon = numpy.mod(ac0_dlon + 180.0, 360.0) - 180.0
    ac0_dlon_trend: numpy.typing.ScalarType = numpy.sum(numpy.sign(ac0_dlon))

    # Make sure that longitude has a monotonic variation along the first
    # row (i.e. "topmost" acrosstrack)
    lon0 = numpy.mod(lon[0, 0] + 180.0, 360.0) - 180.0
    dlon0 = lon[0, :] - lon0
    if 0 < ac0_dlon_trend:
        discordant_ind = numpy.where(dlon0 < 0)
        dlon0[discordant_ind] = numpy.mod(dlon0[discordant_ind] + 360.0, 360.0)
    else:
        discordant_ind = numpy.where(dlon0 > 0)
        dlon0[discordant_ind] = -1.0 * numpy.mod(360.0 - dlon0[discordant_ind],
                                                 360.0)
    lon[0, :] = dlon0 + lon0

    # Adapt each column so that their longitudes are in the same range as
    # values from the first line
    nac = numpy.shape(lon)[1]
    for i in range(0, nac):
        al_lon0 = lon[0, i]
        al_dlon = lon[1:, i] - lon[:-1, i]
        al_dlon, precision_loss = mod360(al_dlon, decimals)
        trunc_to_0 = trunc_to_0 or precision_loss
        al_dlon = numpy.insert(al_dlon, 0, 0.0)
        al_dlon = numpy.mod(al_dlon + 180.0, 360.0) - 180.0
        # TODO (if we find a case where it is necessary): compare sign of
        # ac_dlon with the alongtrack trend
        al_dlon_rel = numpy.cumsum(al_dlon)
        lon[:, i] = al_lon0 + al_dlon_rel

    if trunc_to_0 is True:
        logger.warning('Some angle or longitude values very close to 0 degrees'
                       ' have been truncated to 0 to avoid underflow errors')

    return lon


def fix_swath_longitudes(input_opts: InputOptions,
                         output_opts: OutputOptions,
                         granule: Granule,
                         *args: typing.Iterable[typing.Any],
                         targets: typing.Iterable[str],
                         decimals: int) -> FormatterResult:
    """"""
    _lon = granule.vars['lon']['array']

    lon = _fix_swath_longitudes(_lon, decimals)

    granule.vars['lon']['array'] = lon

    return (input_opts, output_opts, granule)
